# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

#
#	Workaround for Python 2.6/2.7 features
#
try:
	from collections import OrderedDict
except:
	from ordereddict import OrderedDict

import collections 

class OrderedSet(OrderedDict, collections.MutableSet):

    def update(self, *args, **kwargs):
        if kwargs:
            raise TypeError("update() takes no keyword arguments")

        for s in args:
            for e in s:
                 self.add(e)

    def add(self, elem):
        self[elem] = None

    def discard(self, elem):
        self.pop(elem, None)

    def __le__(self, other):
        return all(e in other for e in self)

    def __lt__(self, other):
        return self <= other and self != other

    def __ge__(self, other):
        return all(e in self for e in other)

    def __gt__(self, other):
        return self >= other and self != other

    def __repr__(self):
        return 'OrderedSet([%s])' % (', '.join(map(repr, self.keys())))

    def __str__(self):
        return '{%s}' % (', '.join(map(repr, self.keys())))

    difference = property(lambda self: self.__sub__)
    difference_update = property(lambda self: self.__isub__)
    intersection = property(lambda self: self.__and__)
    intersection_update = property(lambda self: self.__iand__)
    issubset = property(lambda self: self.__le__)
    issuperset = property(lambda self: self.__ge__)
    symmetric_difference = property(lambda self: self.__xor__)
    symmetric_difference_update = property(lambda self: self.__ixor__)
    union = property(lambda self: self.__or__)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

class AbstractService(object):
    def __init__(self, config):
        self._config = config

    def initialize(self):
        pass



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Pong(Peer):
    #Receiving pillows
    Pong = Pillow.InOut

    def __init__(self, *args):
        Peer.__init__(self, *args)
        self._catch(Pong.In.Pong, self.message)

    def message(self, pillow, feathers):
        self._throw(Pong.Out.Pong, int(feathers)+1);



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from database import Database

class ReplicateDatabase(Peer):

    Sending = [
        Database.In.SaveDocument
    ]

    Receiving = [
        Database.In.SaveDocument,
        Database.Out.RequestedDocument
    ]

    def __init__(self, srcRoom, dstRoom):
        Peer.__init__(self, srcRoom)

        self._dstRoom = House.get(dstRoom)

        self._catch(Database.In.SaveDocument, self._saveInDst)
        self._catch(Database.Out.Document, self._replicate)

    def _saveInDst(self, pillow, feathers):
        self._dstRoom.throw(pillow, feathers, me=self)

    def _replicate(self, pillow, feathers):
        self._dstRoom.throw(Database.In.SaveDocument, (feathers, None), me=self)



# -*- coding: utf-8 -*-

# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from twisted.python import log
from twisted.internet import defer, task
import sys, traceback, os, re, json, glob, fnmatch
import socket
from platform import *

logTrace    = False
logDebug    = True
shutdown    = False
config      = None

pLogo       = None
app         = None
appModule   = None
appPath     = None

DEBUG       = 1
TRACE       = 2
WARN        = 3
CRIT        = 4
INFO        = 5

uname     = uname()[0]

widgets  = []

shutdownCBs = []

def replaceValue(dct, val):
    if isinstance(val, dict) and '_path' in dct:
        from wallaby.common.pathHelper import PathHelper
        val = PathHelper.getValue(val, dct['_path'])

    if val == None: 
        val = "-"

    if isinstance(val, (bool, int, float)):
        val = str(val)

    if val in dct:
        val = dct[val]
    elif "*" in dct:
        val = dct["*"].replace("*", val)

    return val

def renderType(val, type, args, edit=False, fct=replaceValue):
    if type == 'datetime':
        if val != None and len(val) >= 5:
            val = ('%02d.%02d.%04d %02d:%02d' % (val[2], val[1], val[0], val[3], val[4]))
        else:
            val = "-"
    if type == 'date':
        if val != None and len(val) >= 3:
            val = ('%02d.%02d.%04d' % (val[2], val[1], val[0]))
        else:
            val = "-"
    elif isinstance(type, (str, unicode)) and type in ('markdown'):
        import markdown
        val = markdown.markdown(val)
    elif isinstance(type, (str, unicode)) and type in ('number' or 'numberedit'):
        if edit:
            try:
                val = unicode(float(val))
            except:
                val = "0.0"
        else:
            try:
                if len(args) > 0:
                    val = ("%." + args[0] + "f") % float(val)
                else:
                    val = ("%d") % int(val)
            except:
                val = "0"

    elif isinstance(type, (str, unicode)) and type in ('currency' or 'currencyedit'):
        if edit:
            try:
                val = unicode(float(val))
            except:
                val = "0.0"
        else:
            try:
                val = u"€ %.2f" % float(val)
            except:
                val = "-"
    elif isinstance(type, dict):
        if val == None:
            val = "-"
        if isinstance(val, list):
            lst = []
            for v in val:
                lst.append(fct(type, v))

            return lst
        else:
            val = fct(type, val)

    if isinstance(val, (dict, list)):
        if isinstance(type, (str, unicode)) and type in ('stringedit', 'autoedit', 'dictedit', 'listedit'):
            import json
            return json.dumps(val)
        else:
            return unicode(len(val)) + u" items"
    return val

def convertType(t):
    if t == None: return t
    if '{' in t:
        try:
            import json
            t = json.loads(t)
        except Exception as e:
            print 'ERROR - Json parse error', str(e), 'in', t

    return t

def extractType(lst, json=True):
    firstList  = []
    secondList = []
    thirdList  = []


    for e in lst:
        m = re.match(r'^<([^>]*)>(.*)', e)
        if m != None:
            type = m.group(1)

            args = []

            if '{' in type:
                if json:
                    try:
                        type = json.loads(type)
                    except Exception as e:
                        print 'ERROR - Json parse error', str(e), 'in', type

            elif ',' in type:
                args = type.split(',')
                type = args.pop(0)

            firstList.append(type)
            secondList.append(args)
            thirdList.append(m.group(2))
        else:
            firstList.append('string')
            secondList.append([])
            thirdList.append(e)

    return firstList, secondList, thirdList

def splitList(lst, sep):
    firstList = []
    secondList = []

    for e in lst:
        if e is None:
            continue

        t = e.rsplit(sep, 1)
        first = t[0]
        if len(t) > 1:
            second = t[1]
        else:
            second = first

        firstList.append(first)
        secondList.append(second)

    return firstList, secondList

def addShutdownCB(cb):
    shutdownCBs.append(cb)

def callShutdownCBs():
    for cb in shutdownCBs:
        cb()

def Property(function):
    keys = 'fget', 'fset', 'fdel'
    func_locals = {'doc':function.__doc__}
    def probeFunc(frame, event, arg):
        if event == 'return':
            locals = frame.f_locals
            func_locals.update(dict((k,locals.get(k)) for k in keys))
            sys.settrace(None)
        return probeFunc
    sys.settrace(probeFunc)
    function()
    return property(**func_locals)

def packagePath(package):
    pathes = []
    try:
        pkg = __import__(package, globals(), locals(), ["*"], 0)
        for p in pkg.__path__: pathes.append(p)  
    except:
        pass 

    return pathes

def imp(name, trace=False):
    m = re.match(".*?\.?(wallaby.*)", name)
    if m:
        name = m.group(1)

    try:
        mod = __import__(name)
        components = name.split('.')
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod
    except Exception as e:
        if trace or not isinstance(e, ImportError):
            import traceback 
            traceback.print_exc(file=sys.stdout)
        return None

def info(*msg):
    log.msg(toMessage("INFO" ,msg), logLevel=INFO)
    
def warn(*msg):    
    log.msg(toMessage("WARNING" ,msg), logLevel=WARN)
        
def crit(*msg):    
    log.err(toMessage("CRITICAL" ,msg), logLevel=CRIT)
        
def debug(*msg):    
    print toMessage("DEBUG", msg) #TODO: remove
    if not logDebug:
        return

    log.msg(toMessage("DEBUG" , msg), logLevel=DEBUG)
    
def trace(*params):
    if not logTrace:
        return

    log.msg(toMessage("TRACE" , params), logLevel=TRACE)
    
    type, value, trace = sys.exc_info()   
    stack = traceback.extract_stack()
    call  = stack[len(stack)-2]
    msg   = "FILE: " + call[0] + "  LINE: " + str(call[1])  
    log.msg(msg, logLevel=TRACE)

def toMessage(prefix, list):
    msg = prefix
    for m in list:
        msg = msg + " " + str(m)
    return msg
    
def deferredSleep(seconds):
    d = defer.Deferred()

    from twisted.internet import reactor
    reactor.callLater(seconds, d.callback, seconds)
    return d

def splitPath(path):
    return path.split('.');

def deferredCall(*args):
    from twisted.internet import reactor
    return task.deferLater(reactor, 0, *args)

## Code borrowed from wxPython's setup and config files
## Thanks to Robin Dunn for the suggestion.
## I am not 100% sure what's going on, but it works!
def opj(*args):
    path = os.path.join(*args)
    return os.path.normpath(path)

def is_package(path):
    return (
        os.path.isdir(path) and
        os.path.isfile(os.path.join(path, '__init__.py'))
        )

def find_data_files(prefix, srcdir, *wildcards, **kw):
    # get a list of all files under the srcdir matching wildcards,
    # returned in a format to be used for install_data
    def walk_helper(arg, dirname, files):
        names = []
        lst, wildcards = arg
        for wc in wildcards:
            wc_name = opj(dirname, wc)
            for f in files:
                filename = opj(dirname, f)

                if fnmatch.fnmatch(filename, wc_name) and not os.path.isdir(filename):
                    names.append(filename)
        if names:
            lst.append( (prefix + "/" + dirname, names ) )

    file_list = []
    recursive = kw.get('recursive', True)
    if recursive:
        os.path.walk(srcdir, walk_helper, (file_list, wildcards))
    else:
        walk_helper((file_list, wildcards),
                    srcdir,
                    [os.path.basename(f) for f in glob.glob(opj(srcdir, '*'))])
    return file_list

def wallabyPackages(packages):
    lst = []
    dirs = {}
    s = set(packages)

    import wallaby as w
    for path in w.__path__:
        packages = find_packages(path, base="wallaby")     
        for p in packages:
            if not '.apps.' in p or 'inspector' in p:
                s.add(p) 
                dirs[p] = packages[p]

    for p in s: lst.append(p)
    return lst, dirs

def find_packages(path, base="" ):
    """ Find all packages in path """
    packages = {}
    for item in os.listdir(path):
        dir = os.path.join(path, item)
        if is_package( dir ):
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item

            packages[module_name] = dir
            packages.update(find_packages(dir, module_name))
    return packages




# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from editor import Editor

class BooleanEditor(Editor):
    def _fieldChanged(self):
        value = False
        if self._valueCallback:
            value = self._valueCallback()
        if (self._document):
            self._document.set(self._path, value) 

        self._throwFieldChanged()



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from uuid import uuid4
import wallaby.FX as FX
from wallaby.common.pathHelper import PathHelper

class Document:
    def __init__(self, documentID=None, data=None):
        if data == None:
            if documentID == None:
                documentID = uuid4().hex
            data = {'_id':documentID}
        else:
            if documentID != None:
                data['_id'] = documentID
            elif '_id' in data:
                documentID = data['_id']
            else:
                documentID = uuid4().hex
                data['_id'] = documentID

        self._data = data
        self.documentID = documentID
        self.selection = {}

    def __str__(self):
        # return "<"+self.__class__.__name__+"("+hex(id(self))+") with id: '"+self.documentID+"'>"
        if self.documentID != None:
            if len(self.documentID) == 32:
                return "<"+self.__class__.__name__+"(..."+self.documentID[28:32]+")>"
            return "<"+self.__class__.__name__+"("+self.documentID+")>"
        return "<"+self.__class__.__name__+"(None)>"

    def hasConflicts(self):
        return False

    def getConflicts(self):
        return None

    def changes(self, doc):
        return []

    def getFlat(self, path):
        if path not in self._data:
            return None
        else:
            return self._data[path]

    def get(self, path, **kw):
        return PathHelper.getValue(self._data, path, self.selection, **kw)

    def getSelection(self, path):
        return PathHelper.getSelection(path, self.selection)

    def sub(self, key, path=None, dct=None):
        if dct == None: dct = {} 

        if isinstance(key, dict):
            for k,v in key.items():
                subpath = k
                if path != None: subpath = path + "." + k

                self.sub(v, subpath, dct)
        else:
            PathHelper.setValue(dct, path, PathHelper.getValue(self._data, path))

        return dct

    def getDictionary(self):
        dict = copy.copy(self._data)
        if '_id' in dict:
            del dict['_id']
        if '_rev' in dict:
            del dict['_rev']
        if 'wallabyUser' in dict:
            del dict['wallabyUser']

        return dict

    def rev(self):
        return self._data.get("_rev", None)

    def set(self, path, value):
        return PathHelper.setValue(self._data, path, value, self.selection)

    def overwrite(self, path, value):
        return PathHelper.setValue(self._data, path, value, self.selection, recurse=True)

    def merge(self, path, value):
        return PathHelper.setValue(self._data, path, value, self.selection, merge=True)


    def select(self, path, index):
        PathHelper.setValue(self.selection, path + "._selection", index)

    def solveConflicts(self, document):
        pass

    def __deepcopy__(self, memo):
        doc = Document()
        doc.documentID = self.documentID
        doc.selection = self.selection

        import copy
        doc._data = copy.deepcopy(self._data)
        return doc

    def clone(self):
        doc = Document()
        doc.documentID = self.documentID
        doc.selection = self.selection
        doc._data = copy.deepcopy(self._data)
        return doc

    def resetDocumentID(self, id=None):
        if id == None:
            id = uuid4().hex

        self.documentID = id
        self._data['_id'] = id
        if '_rev' in self._data:
            del self._data['_rev']

    def deferredGetAttachment(self, name): #TODO: Not really deferred?
        # Hack for dummy images
        try:
            import os
            f = open('./tmp/' + name)
            size = os.fstat(f.fileno()).st_size
        except:
            return None, 0

    def setAttachment(self, name, attachment):
        f = open('./tmp/' + name, "w")
        f.write(attachment)
        f.close()

    def deleteAttachment(self, name):
        pass

    @FX.Property                                                                                                                                      
    def documentID():
        doc = "The documentID Property."
        def fget(self):
            return self._documentID
        def fset(self, value):
            self._documentID = value

    @FX.Property                                                                                                                                      
    def selection():
        doc = "The selection Property."
        def fget(self):
            return self._selection
        def fset(self, value):
            self._selection = value



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from roomDebugger import RoomDebugger
from documentChanger import DocumentChanger
# from credentials import Credentials
from listPeer import ListPeer
from viewer import Viewer
from editor import Editor

class RoomDebuggerRouting(Peer):
   
    Sending = [
        Viewer.In.Document,
        Editor.In.Enable,
        RoomDebugger.In.RequestDebugInfo,
        ListPeer.In.RequestedDocument
    ]

    Receiving = [
        RoomDebugger.Out.RequestedRooms,
        Editor.Out.FieldChanged
    ]

    Routings = [
        (RoomDebugger.Out.RequestedRooms, ListPeer.In.RequestedDocument),
        (RoomDebugger.Out.RequestedRooms, Viewer.In.Document),
        (ListPeer.Out.RequestDocument, RoomDebugger.In.RequestRooms),
        (RoomDebugger.Out.RequestedDebugInfo, Viewer.In.Document),
        (DocumentChanger.Out.SelectionChanged, Viewer.In.Refresh)
    ]

    # FIXME: (RoomDebugger.Out.RequestedActions, Viewer.In.Document),

    def __init__(self, room):
        Peer.__init__(self, room)

        self._roomDocument = None

        self._catch(Editor.Out.FieldChanged, self._fieldChanged)
        self._catch(RoomDebugger.Out.RequestedRooms, self._rooms)

    def initialize(self):
        # FIXME: self._throw(Credentials.Out.Credential, Document())
        self._throw(Editor.In.Enable, True)

    def _fieldChanged(self, pillow, path):
        if path == 'room':
            if self._roomDocument != None:
                room = self._roomDocument.get(path)

            self._throw(RoomDebugger.In.RequestDebugInfo, room)

    def _rooms(self, pillow, document):
        self._roomDocument = document 
        self._throw(ListPeer.In.RequestedDocument, document)
        self._throw(Viewer.In.Document, document)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Stack(Peer):
    ShowLayer = Pillow.In
    HideLayer = Pillow.In
    EnableLayer = Pillow.In
    DisableLayer = Pillow.In

    def __init__(self, room, controller):
        Peer.__init__(self, room)
        self._controller = controller
        self._catch(Stack.In.ShowLayer, self._showLayer)
        self._catch(Stack.In.HideLayer, self._hideLayer)
        self._catch(Stack.In.EnableLayer, self._enableLayer)
        self._catch(Stack.In.DisableLayer, self._disableLayer)

    def _showLayer(self, pillow, name):
        self._controller.showLayer(name)

    def _hideLayer(self, pillow, name):
        self._controller.hideLayer(name)

    def _enableLayer(self, pillow, name):
        self._controller.enableLayer(name)

    def _disableLayer(self, pillow, name):
        self._controller.disableLayer(name)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from viewer import *
from editor import *
import copy

from wallaby.common.document import *

class History(Peer):
    Back = Pillow.In
    Forward = Pillow.In

    Receiving = [
        Viewer.In.Document,
        Editor.Out.FieldChanged
    ]

    Sending = [
        Editor.Out.FieldChanged
    ]

    def __init__(self, name):
        Peer.__init__(self, name)

        self._queue = []
        self._fqueue = []
        self._doc = None
        self._lastDoc = None

        self._ignore = False
   
        self._catch(Viewer.In.Document, self._newDoc)
        self._catch(Editor.Out.FieldChanged, self._change)
        self._catch(History.In.Back, self._back)
        self._catch(History.In.Forward, self._forward)

    def _back(self, *args):
        if len(self._queue) == 0: return

        path, value = self._queue.pop()        

        oldVal = self._doc.get(path)
        self._fqueue.append( (path, oldVal) )

        self._lastDoc = Document(data=copy.deepcopy(self._doc._data))

        self._doc.set(path, value)
        self._ignore = True
        self._throw(Editor.Out.FieldChanged, path)

    def _forward(self, *args):
        if len(self._fqueue) == 0: return

        path, value = self._fqueue.pop()
        oldVal = self._doc.get(path)
        self._queue.append( (path, oldVal) )

        self._lastDoc = Document(data=copy.deepcopy(self._doc._data))

        self._doc.set(path, value)
        self._ignore = True
        self._throw(Editor.Out.FieldChanged, path)

    def _newDoc(self, pillow, doc):
        self._doc = doc
        self._lastDoc = Document(data=copy.deepcopy(self._doc._data))

    def _change(self, pillow, path):
        if None in (self._doc, self._lastDoc): return

        if self._ignore: 
            self._ignore = False
            return

        self._queue.append( (path, self._lastDoc.get(path)) )
        self._lastDoc.set(path, self._doc.get(path))

        self._fqueue = []



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Debugger(Peer):
    from wallaby.pf.room import Room 

    Receiving = [
        Room.Out.NoHandlerFound
    ]

    def __init__(self, room, pillow=Room.Out.NoHandlerFound, notFoundLimit=5):
        Peer.__init__(self, room)
        self._roomName = room
        self._catch(pillow, self._message, passThrower=True)
        self._notFoundLimit = notFoundLimit

        self._notFound = {}

    def color(self, this_color, string):
        return "\033[" + this_color + "m" + string + "\033[0m"

    def _message(self, pillow, feather, thrower=None):
        from wallaby.pf.room import Room
        if self._notFoundLimit > 0 and pillow == Room.Out.NoHandlerFound and feather in self._notFound and self._notFound[feather] == self._notFoundLimit: return

        thrower = str(thrower)
        if '.' in thrower:
            import re
            match = re.match('<([^\s]+)', thrower)
            thrower = match.group(1).split('.')
            thrower = thrower[-2] + "." + thrower[-1]

        out = "|{0:<16}|{1:<48}|{2:<64}|{3}".format(self._roomName,thrower,pillow,feather)

        if pillow == Room.Out.NoHandlerFound:
            if feather not in self._notFound:
                self._notFound[feather] = 1
            else:
                self._notFound[feather] += 1

            if self._notFound[feather] == self._notFoundLimit:
                out += "..."

            out = self.color('1;31', out)

        import wallaby.FX as FX
        FX.debug(out)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import Peer

class RouteAction(Peer):
    
    def __init__(self, srcRoom, dstRoom, pillow):
        Peer.__init__(self, srcRoom)
        
        from wallaby.pf.room import House
        self._dstRoom = House.get(dstRoom)

        self._catch(pillow, self._message)

    def _message(self, *args):
        self._dstRoom.throw(*args)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

class ConfigInterface(object):
    def getConfig(self, key=None, type=None): pass
    def setConfig(self, value, key=None, type=None): pass
    def resetConfig(self, key=None, type=None): pass



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from wallaby.common.document import Document
from documentChanger import DocumentChanger

class DocumentCache(Peer):
    from wallaby.pf.pillow import Pillow

    Description = "Provides caching and shadow copy mechanism for documents"

    Suggests = [
        "DocumentDatabase"
    ]

    #Receiving Pillows
    ForgetDocument = Pillow.In
    DeleteShadowCopy = Pillow.In
    CreateShadowCopy = Pillow.In
    ReplaceShadowCopy = Pillow.In
    ChangesInShadowCopy = Pillow.In
    StoreShadowCopy = Pillow.In
    StoreShadowCopyAs = Pillow.In

    RequestDocument = Pillow.InOut
    DocumentChanged = Pillow.InOut
    DeleteDocument = Pillow.InOut
    DocumentSaved = Pillow.InOut
    CreateDocument = Pillow.InOut
    CreateAndSaveDocument = Pillow.InOut
    DocumentCreated = Pillow.InOut
    RequestedDocument = Pillow.InOut

    #Sending Pillows
    SaveDocument = Pillow.Out
    ShadowCopyAlreadyExists = Pillow.Out
    DoesNotExist = Pillow.Out
    ShadowCopyNotFound = Pillow.Out

    def __init__(self, *args):
        Peer.__init__(self, *args)
        self._documentStore = {} # id -> document
        self._shadowCopyStore = {} # id -> document
        self._requestedDocuments = [] # id

        self._catch(DocumentCache.In.DocumentChanged, self._documentChanged)
        self._catch(DocumentCache.In.CreateDocument, self._createDocument)
        self._catch(DocumentCache.In.CreateAndSaveDocument, self._createAndSaveDocument)
        self._catch(DocumentCache.In.DocumentCreated, self._documentCreated)
        self._catch(DocumentCache.In.RequestedDocument, self._documentFromDatabase)
        self._catch(DocumentCache.In.RequestDocument, self._requestDocument)
        self._catch(DocumentCache.In.ForgetDocument, self._forgetDocument)
        self._catch(DocumentCache.In.DeleteDocument, self._deleteDocument)
        self._catch(DocumentCache.In.DeleteShadowCopy, self._deleteShadowCopy)
        self._catch(DocumentCache.In.CreateShadowCopy, self._createShadowCopy)
        self._catch(DocumentCache.In.ChangesInShadowCopy, self._changesInShadowCopy)
        self._catch(DocumentCache.In.ReplaceShadowCopy, self._replaceShadowCopy)
        self._catch(DocumentCache.In.StoreShadowCopy, self._storeShadowCopy)
        self._catch(DocumentCache.In.StoreShadowCopyAs, self._storeShadowCopyAs)
        self._catch(DocumentCache.In.DocumentSaved, self._documentSaved)

    def _documentChanged(self, pillow, (documentID, rev) ):
        if documentID in self._documentStore:
            if rev == self._documentStore[documentID].rev():
                # ignore already known revision
                return

            self._throw(DocumentCache.Out.RequestDocument, documentID)

    def _documentCreated(self, pillow, doc):
        self._shadowCopyStore[doc.documentID] = doc
        self._throw(DocumentCache.Out.DocumentCreated, doc)

    @defer.inlineCallbacks
    def _createDocument(self, pillow, doc):
        if isinstance(doc, str) or isinstance(doc, unicode):
            try:
                doc = yield DocumentChanger.transform(None, doc)
            except Exception as e:
                doc = None
                print "ERROR - Exception: " + str(e)

        if type(doc) is dict:
            self._throw(DocumentCache.Out.CreateDocument, doc)
        else:
            self._throw(DocumentCache.Out.CreateDocument, None)

    @defer.inlineCallbacks
    def _createAndSaveDocument(self, pillow, doc):
        if isinstance(doc, str) or isinstance(doc, unicode):
            try:
                doc = yield DocumentChanger.transform(None, doc)
            except Exception as e:
                doc = None
                print "ERROR - Exception: " + str(e)

        if type(doc) is dict:
            self._throw(DocumentCache.Out.CreateAndSaveDocument, doc)
        else:
            self._throw(DocumentCache.Out.CreateAndSaveDocument, None)

    def _documentFromDatabase(self, pillow, document):
        if document == None: return
            
        if document.documentID in self._documentStore:
            if document.rev() == self._documentStore[document.documentID].rev(): 
                # Ignore same revision
                return

            document.selection = self._documentStore[document.documentID].selection
            self._documentStore[document.documentID]=document
            self._throw(DocumentCache.Out.DocumentChanged, document)
        elif document.documentID in self._requestedDocuments:
            self._documentStore[document.documentID]=document
            self._requestedDocuments.remove(document.documentID)
            self._requestDocument(None, document.documentID)

    def _requestDocument(self, pillow, documentID):
        if documentID in self._documentStore:
            self._throw(DocumentCache.Out.RequestedDocument, self._documentStore[documentID])
        elif documentID in self._shadowCopyStore:
            self._throw(DocumentCache.Out.RequestedDocument, self._shadowCopyStore[documentID])
        elif documentID not in self._requestedDocuments:
            self._requestedDocuments.append(documentID)
            self._throw(DocumentCache.Out.RequestDocument, documentID)

    def _forgetDocument(self, pillow, documentID):
        if documentID in self._documentStore:
            del self._documentStore[documentID]
        if documentID in self._shadowCopyStore:
            del self._shadowCopyStore[documentID]
        if documentID in self._requestedDocuments:
            self._requestedDocuments.remove(documentID)

    def _deleteDocument(self, pillow, documentID):
        if documentID in self._documentStore:
            document = self._documentStore[documentID]
            self._forgetDocument(pillow, documentID)
            self._throw(DocumentCache.Out.DeleteDocument, document)
        else:
            self._throw(DocumentCache.Out.DoesNotExist, documentID)

    def _deleteShadowCopy(self, pillow, documentID):
        if documentID not in self._shadowCopyStore:
            self._throw(DocumentCache.Out.ShadowCopyNotFound, documentID)
        else:
            del self._shadowCopyStore[documentID]

    def _replaceShadowCopy(self, pillow, documentID):
        self._deleteShadowCopy(pillow, documentID)
        self._createShadowCopy(pillow, documentID)

    def _createShadowCopy(self, pillow, documentID):
        if documentID not in self._documentStore:
            self._throw(DocumentCache.Out.DoesNotExist, documentID)
        elif documentID in self._shadowCopyStore:
            self._throw(DocumentCache.Out.ShadowCopyAlreadyExists, documentID)
        else:
            doc = self._documentStore[documentID]

            import copy
            self._documentStore[documentID] = copy.deepcopy(doc)
            self._shadowCopyStore[documentID] = doc

    def _changesInShadowCopy(self, pillow, documentID):
        if documentID not in self._documentStore:
            self._throw(DocumentCache.Out.DoesNotExist, documentID)
        else:
            self._documentStore[documentID].changes(self._shadowCopyStore[documentID])
            pass

    def _storeShadowCopyAs(self, pillow, (documentID, newID)):
        if documentID not in self._shadowCopyStore and documentID in self._documentStore:
            self._shadowCopyStore[documentID] = self._documentStore[documentID]  

        if documentID not in self._shadowCopyStore:
            self._throw(DocumentCache.Out.ShadowCopyNotFound, documentID)
        else:
            shadowCopy  = self._shadowCopyStore[documentID]

            shadowCopy = shadowCopy.clone()
            shadowCopy.resetDocumentID(newID)
            documentID = newID

            self._shadowCopyStore[documentID] = shadowCopy

            self._throw(DocumentCache.Out.SaveDocument, (shadowCopy, None) )

    def _storeShadowCopy(self, pillow, documentID):
        if documentID not in self._shadowCopyStore:
            self._throw(DocumentCache.Out.ShadowCopyNotFound, documentID)
        else:
            shadowCopy  = self._shadowCopyStore[documentID]
            oldDocument = None

            if documentID in self._documentStore:
                oldDocument = self._documentStore[documentID]
                shadowCopy.solveConflicts(oldDocument)

            # Update changed document in document store.
            # Needed for Databases without changes stream! 
            self._documentStore[documentID] = shadowCopy

            self._throw(DocumentCache.Out.SaveDocument, (shadowCopy, oldDocument) )

    def _documentSaved(self, pillow, documentID):
        if documentID in self._shadowCopyStore:
            self._deleteShadowCopy(None, documentID)

        self._throw(DocumentCache.Out.DocumentSaved, documentID)




# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from editDocument import EditDocument
from viewer import Viewer

class Relation(Peer):
    from abstractQuery import AbstractQuery

    Add = Pillow.In
    AddAndSave = Pillow.In

    Sending = [
        EditDocument.In.Rollback, 
        Viewer.In.Document,
        AbstractQuery.In.Query
    ]

    Receiving = [
        Viewer.In.Document,
    ]

    def __init__(self, srcRoom, dstRoom=None, viewIdentifier=None):
        Peer.__init__(self, srcRoom)

        from wallaby.pf.room import House
        self._dstRoom = House.get(dstRoom)

        self._viewIdentifier = viewIdentifier
        self._documentID = None
        self._queryDoc = None
        self._queryRoom = None

        self._catch(Viewer.In.Document, self._docChanged)
        self._catch(Relation.In.Add, self._add)
        self._catch(Relation.In.AddAndSave, self._addAndSave)

        if viewIdentifier != None:
            self._queryRoom = House.get(self._viewIdentifier.upper())

        if self._queryRoom:
            self._queryRoom.catch(Viewer.In.Document, self._queryDocChanged)


    def rooms(self, pillow, feathers):
        rooms = Peer.rooms(self)

        if pillow in (Relation.In.Add, Relation.In.AddAndSave):
            if ';' in feathers:
                dstRoom, relationKey, doc = feathers.split(';')
                if dstRoom == self._dstRoom._name:
                    rooms.add(self._dstRoom._name)

        return rooms

    def destroy(self, remove=False):
        Peer.destroy(self, remove=False)
        if self._queryRoom: self._queryRoom.catch(Viewer.In.Document, self._queryDocChanged)

    def _add(self, pillow, params):
        self._new(params, EditDocument.In.New)

    def _addAndSave(self, pillow, params):
        self._new(params, EditDocument.In.NewAndSave)

    def _new(self, params, sendPillow):
        if self._documentID:
            doc = {}
            dstRoom = self._dstRoom._name
            relationKey = params

            if ';' in params:
                dstRoom, relationKey, doc = params.split(';')
                #print dstRoom, relationKey, doc
                try:
                    import json
                    doc = json.loads(doc)
                except Exception as e:
                    print "JSON ERROR (RELATION)", e
                    doc = {}

            if dstRoom != self._dstRoom._name:
                return

            doc[relationKey] = self._documentID

            print "Try to create new doc", doc, "in room", self._dstRoom
            self._dstRoom.throw(sendPillow, doc, me=self)

    def _queryDocChanged(self, pillow, doc):
        self._queryDoc = doc
        self._refresh(pillow, None)

    def _docChanged(self, pillow, doc):
        if doc == None:
            self._documentID = None
        else:
            self._documentID = doc.documentID

        # self._dstRoom.throw(EditDocument.In.Rollback, None, me=self)
        # self._dstRoom.throw(Viewer.In.Document, None, me=self)
        self._refresh(pillow, None)

    def _refresh(self, pillow, feathers):
        if self._documentID != None and self._queryDoc:
            desc = self._queryDoc.get('args.descending')
            
            if desc:
                self._queryDoc.set('args.endkey', [self._documentID])
                self._queryDoc.set('args.startkey', [self._documentID, {}])
            else:
                self._queryDoc.set('args.startkey', [self._documentID])
                self._queryDoc.set('args.endkey', [self._documentID, {}])


            from multiViewer import MultiViewer
            self._dstRoom.throw(MultiViewer.In.Refresh, None)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Enabler(Peer):
    Enable = Pillow.InState

    def __init__(self, room, cb, feathers):
        Peer.__init__(self, room)
        self._cb = cb
        self._feathers = {}
        self._otherFeathers = {}

        self._enabled = {}

        self._negate = False
        self._negateOther = {}

        if feathers is None: return

        for p in feathers:
            if p is None: continue
            if ':' in p:
                room, p = p.split(':')
                if room not in self._otherFeathers: self._otherFeathers[room] = {}

                if p.startswith('!'):
                    p = p[1:]
                    self._negateOther[room] = True
                else:
                    self._negateOther[room] = False

                self._otherFeathers[room][p] = True
            else:
                if p.startswith('!'): 
                    p = p[1:]
                    self._negate = True

                self._feathers[p] = True

        if len(self._otherFeathers) > 0:
            if len(self._feathers) > 0:
                self._otherFeathers[self._roomName] = self._feathers
                self._negateOther[self._roomName] = self._negate

            i = 0
            for roomName in self._otherFeathers.keys():
                self._enabled[roomName] = False
                from wallaby.pf.room import House
                room = House.get(roomName)

                from functools import partial
                room.catch(Enabler.In.Enable, partial(self._enableOther, roomName))

        else:
            self._catch(Enabler.In.Enable, self._enable)

    # TODO: Was passiert hier? Warum?
    def _enableOther(self, room, pillow, feather):
        if feather in self._otherFeathers[room]:
            self._enabled[room] = not self._negateOther[room]
        else:
            self._enabled[room] = self._negateOther[room]

        enabled = True

        for k, v in self._enabled.items():
            if not v:
                enabled = False

        self._cb(enabled)

        # print "EnableOther", room, pillow, feather, enabled, self._negateOther[room]

    def _enable(self, pillow, feather):
        if feather in self._feathers:
            self._cb(not self._negate)
        else:
            self._cb(self._negate)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

class Pillow(type):
    Out        = "WLBY_API_OUT"
    OutState   = "WLBY_API_OUT_STATE"
    In         = "WLBY_API_IN"
    InState    = "WLBY_API_IN_STATE"
    InOut      = "WLBY_API_IN_OUT"
    InOutState = "WLBY_API_IN_OUT_STATE"

    Runtime = "WLBY_API_RUNTIME"
    Startup = "WLBY_API_STARTUP"

    def __new__(mcs, name, bases, dict):
        in_dict = {}
        out_dict = {}

        inPillows = []
        outPillows = []

        import copy
        dct = copy.copy(dict)

        for key, val in dct.items():
            if val == Pillow.InOut: 
                in_dict[key] = name + ".In." + key 
                out_dict[key] = name + ".Out." + key 
                inPillows.append(key)
                outPillows.append(key)
            elif val == Pillow.InOutState: 
                in_dict[key] = name + ".In." + key + "!"
                out_dict[key] = name + ".Out." + key + "!"
                inPillows.append(key)
                outPillows.append(key)
            elif val == Pillow.In: 
                in_dict[key] = name + ".In." + key 
                inPillows.append(key)
            elif val == Pillow.InState: 
                in_dict[key] = name + ".In." + key  + "!"
                inPillows.append(key)
            elif val == Pillow.Out: 
                out_dict[key] = name + ".Out." + key 
                outPillows.append(key)
            elif val == Pillow.OutState: 
                out_dict[key] = name + ".Out." + key  + "!"
                outPillows.append(key)
            else:
                continue

            del dict[key]

        if len(in_dict)+len(out_dict) > 0:
            if '__doc__' not in dict:
                dict['__doc__'] = ''
            if len(in_dict) > 0:
                dict['__doc__'] += '\nIngoing Pillows:'
                for key in in_dict:
                    dict['__doc__'] += '\n\t* '+in_dict[key]
            if len(out_dict) > 0:
                dict['__doc__'] += '\nOutgoing Pillows:'
                for key in out_dict:
                    dict['__doc__'] += '\n\t* '+out_dict[key]
            dict['__doc__'] += '\n'

        if not 'Sending' in dict:
            dict['Sending'] = set()
        else:
            dict['Sending'] = set(dict['Sending'])

        if not 'Receiving' in dict:
            dict['Receiving'] = set()
        else:
            dict['Receiving'] = set(dict['Receiving'])

        if 'Routings' in dict:
            for f, t in dict['Routings']:
                dict['Sending'].add(t)
                dict['Receiving'].add(f)
        else:
            dict['Routings'] = set()

        dict['In'] = type(name + 'InPillows', (object,), in_dict)
        dict['Out'] = type(name + 'OutPillows', (object,), out_dict)

        dict['InPillows'] = inPillows 
        dict['OutPillows'] = outPillows 

        if not 'ObjectType' in dict:
            dict['ObjectType'] = Pillow.Startup

        return type.__new__(mcs, name, bases, dict)




# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class UIPeer(Peer):
    def __init__(self, room, path=None):
        Peer.__init__(self, room)
        self._roomName = room
        self._path = path

        from credentialsConsumer import CredentialsConsumer
        self._cc = CredentialsConsumer(self._newCredentials)
        self._credentials = None

    def destroy(self, remove=False):
        Peer.destroy(self, remove)
    
        if self._cc: self._cc.destroy(remove)
        self._cc = None

    def _newCredentials(self, pillow, feathers):
        self._credentials = feathers
        self._refresh(None, self._path)

    def _refresh(self, pillow, path):
        pass



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class ListDocumentCache(Peer):
    from documentCache import DocumentCache
    from listPeer import ListPeer

    Routings = [
        (DocumentCache.Out.RequestedDocument, ListPeer.In.RequestedDocument),
        (DocumentCache.Out.DocumentChanged, ListPeer.In.RequestedDocument),
        (ListPeer.Out.RequestDocument, DocumentCache.In.RequestDocument)
    ]

    def __init__(self, room):
        Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Statistic(Peer):
    #Sending pillows
    MessagesSinceLastReport = Pillow.Out

    def __init__(self, room, pillow="*", delay=1.0):
        Peer.__init__(self, room)
        self._delay = delay
        self._count = 0
        self._catch(pillow, self._message)

        self._initialized = False

    def initialize(self):
        if self._initialized: return
        self._initialized = True

        self._triggerReport()
        import time

    def _triggerReport(self):
        from twisted.internet import reactor, task
        task.deferLater(reactor, self._delay, self._sendReport)

    def _sendReport(self):
        self._throw(Statistic.Out.MessagesSinceLastReport, self._count)
        self._count = 0
        self._triggerReport()

    def _message(self, pillow, feathers):
        self._count += 1



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *
from editDocument import EditDocument

class ConfigParty(Peer):

    def __init__(self, room, database=None, docId=None, inspectorDb=None, dbClass=None, *args, **ka):
        Peer.__init__(self, room)

        from database import Database
        from documentDatabase import DocumentDatabase
        from documentCache import DocumentCache
        from enableDocument import EnableDocument
        from documentChanger import DocumentChanger
        from debugger import Debugger
        from dialog import Dialog

        self._docId = docId
        self._inspectorDb = inspectorDb

        # Peers...
        self.add(Debugger(room))
        self.add(DocumentCache(room))
        self.add(EditDocument(room, ignoreNotFound=True))
        self.add(EnableDocument(room))
        self.add(DocumentChanger(room))

        if dbClass is None:
            # try CouchDB as default
            from couchDB import CouchDB
            self.add(CouchDB(room, query=False, databaseName=database, **ka))
        else:
            self.add(dbClass(room, query=False, databaseName=database, **ka))

        Routings = [
            (DocumentCache.Out.SaveDocument, Database.In.SaveDocument),
            (DocumentCache.Out.CreateDocument, Database.In.CreateDocument),
            (DocumentCache.Out.CreateAndSaveDocument, Database.In.CreateAndSaveDocument),
            (DocumentCache.Out.DeleteDocument, Database.In.DeleteDocument),
            (Database.Out.DocumentChanged, DocumentCache.In.DocumentChanged),
            (Database.Out.NewDocument, DocumentCache.In.DocumentCreated),
            (Database.Out.DocumentSaved, DocumentCache.In.DocumentSaved),
            (Database.Out.DocumentCreated, DocumentCache.In.DocumentSaved),
            (DocumentCache.Out.RequestDocument, Database.In.RequestDocument) 
        ]

        self._catch(Database.Out.DocumentNotFound, self._notFound)

        if inspectorDb != None:
            self._catch(Database.Out.RequestedDocument, self._requestedDocument)
        else:
            Routings.append( (Database.Out.RequestedDocument, DocumentCache.In.RequestedDocument) )

        for routing in Routings: self._addRouting(*routing) 

    def _notFound(self, pillow, docId):
        if docId == self._docId:
            from wallaby.common.document import Document
            doc = Document(self._docId)
            self._requestedDocument(pillow, doc)

    @defer.inlineCallbacks
    def _requestedDocument(self, pillow, doc):
        if doc is not None:
            if doc.documentID == self._docId: 
                if self._inspectorDb != None:
                    inspectorDoc = {}
                    try:
                        import wallaby.backends.couchdb as couch
                        defaultDB = couch.Database.getDatabase()                
                        couch.Database.setURLForDatabase(self._inspectorDb, couch.Database.getURLForDatabase(None)) # default
                
                        user, password = defaultDB.credentials()
                        if user != None and password != None:
                            couch.Database.setLoginForDatabase(self._inspectorDb, user, password)

                        db = couch.Database.getDatabase(self._inspectorDb)
                        print db._url
                        inspectorDoc = yield db.get("WallabyApp2")
                        if inspectorDoc == None: inspectorDoc = None
                    except Exception as e:
                        print "No inspector Doc found", e

                    widgets = {}
                    if inspectorDoc != None and "widgets" in inspectorDoc: widgets = inspectorDoc["widgets"]
                    for k, v in widgets.items(): doc.set("widgets." + k, v)

            from documentCache import DocumentCache
            self._throw(DocumentCache.In.RequestedDocument, doc)

    def initialize(self):
        if self._docId != None:
            self._throw(EditDocument.In.Load, self._docId)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from documentCache import DocumentCache
from documentChanger import DocumentChanger
from database import Database
from viewer import Viewer
from editor import Editor

class DirectEditDocument(Peer):
    Description = ("Provides a direct document edit/view logic", '{"id": null}')

    Save = Pillow.In
    SaveAs = Pillow.In
    Load = Pillow.In

    ChangeID = Pillow.In

    Sending = [
        Viewer.In.Document,
        Viewer.In.Refresh,
        Editor.In.Enable,
        DocumentCache.In.CreateShadowCopy,
        DocumentCache.In.RequestDocument
    ]

    Receiving = [
        DocumentCache.Out.RequestedDocument,
        DocumentCache.Out.DocumentChanged,
        DocumentCache.Out.DocumentSaved
    ]

    Routings = [
        (Editor.Out.FieldChanged, Viewer.In.Refresh)
    ]

    def __init__(self, room, id=None):
        Peer.__init__(self, room)

        self._id = id
        self._newID = id
        self._doc = None
        self._requestNew = False

        self._catch(DirectEditDocument.In.Load, self._load)
        self._catch(DirectEditDocument.In.Save, self._save)
        self._catch(DirectEditDocument.In.SaveAs, self._saveAs)
        self._catch(DirectEditDocument.In.ChangeID, self._changeID)
        self._catch(DocumentCache.Out.RequestedDocument, self._document)
        self._catch(DocumentCache.Out.DocumentChanged, self._changed)
        self._catch(DocumentCache.Out.DocumentSaved, self._saved)

        self._addRouting(DocumentChanger.Out.SelectionChanged, Viewer.In.Refresh)

    def _enable(self, enabled):
        self._throw(Editor.In.Enable, enabled)

    def _updateAll(self):
        self._throw(Viewer.In.Document, self._doc)

    def _changeID(self, pillow, newID):
        self._newID = newID

    def _load(self, pillow, feathers):
        if feathers == None and self._newID != None:
            feathers = self._newID

        if self._id == feathers: return

        self._throw(DocumentCache.In.DeleteShadowCopy, self._id)
        self._id = feathers

        self._enable(False)
        self._throw(DocumentCache.In.RequestDocument, self._id)

    def _saved(self, pillow, feathers):
        if feathers != self._id: return

        if self._requestNew:
            self._requestNew = False
            self._throw(DocumentCache.In.RequestDocument, self._id)
        else:
            self._throw(DocumentCache.In.CreateShadowCopy, self._doc.documentID)
            self._enable(True)

    def _saveAs(self, pillow, newID):
        if newID != None:
            self._newID = newID

        self._id = self._newID

        self._enable(False)

        if self._doc.documentID != self._id:
            self._requestNew = True
            self._throw(DocumentCache.In.StoreShadowCopyAs, (self._doc.documentID, self._id))
        else:
            self._throw(DocumentCache.In.StoreShadowCopy, self._doc.documentID)

    def _save(self, pillow, feathers):
        if self._newID != None:
            from twisted.internet import reactor
            reactor.callLater(0, self._saveAs, pillow, self._newID)
            return

        self._enable(False)
        self._throw(DocumentCache.In.StoreShadowCopy, self._doc.documentID)

    def _changed(self, pillow, doc):
        if doc.documentID != self._id: return 
        self._throw(DocumentCache.In.ReplaceShadowCopy, self._doc.documentID)

        self._updateAll() 

    def _document(self, pillow, doc):
        if doc.documentID != self._id: return 

        self._throw(DocumentCache.In.CreateShadowCopy, doc.documentID)

        self._doc = doc
        self._updateAll()
        self._enable(True)

    def initialize(self):
        self._throw(DocumentCache.In.RequestDocument, self._id)
        self._enable(False)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

# Common imports 
from wallaby.pf.pillow import Pillow
from twisted.internet import defer

class Peer(object):
    from wallaby.pf.pillow import Pillow

    __metaclass__ = Pillow

    peers = []

    def __init__(self, roomName):
        if roomName == None: roomName = "None"

        self.initialized = False

        from wallaby.pf.room import House
        self.__room = House.get(roomName)
        Peer.peers.append(self)

        self._roomName = roomName

        self._routings = {}
        self._peers = []
        self._catchers = set()

        if 'Routings' in self.__class__.__dict__:
            for f, t in self.__class__.Routings:
                self._addRouting(f, t)

        self.destroyed = False
        self.__room.enter(self)

        from wallaby.pf.peer.viewer import Viewer

    def rooms(self, *args):
        from wallaby.common.sets import OrderedSet
        rooms = OrderedSet()
        rooms.add(self._roomName)
        return rooms

    def dynamicPillows(self):
        return []

    @classmethod
    def initializeAll(cls, onlyNew=False):
        for peer in Peer.peers:
            if not peer.initialized or not onlyNew:
                peer.initialize()
                peer.initialized = True

    def destroy(self, remove=False):
        s = set(self._catchers)
        for act in s: self._uncatch(*act)
        for p in self._peers: p.destroy()
        if remove and self in Peer.peers: Peer.peers.remove(self)

        self._catchers = set()

        self.__room.leave(self)
        self.destroyed = True

    def add(self, peer):
        self._peers.append(peer)

    def _addRouting(self, sourcePillow, destinationPillow):
        if not sourcePillow in self._routings:
            self._routings[sourcePillow] = set()
            self._catch(sourcePillow, self._routePillow)

        self._routings[sourcePillow].add(destinationPillow)

    def _routePillow(self, pillow, feathers):
        if pillow in self._routings:
            for routing in self._routings[pillow]:
                self.__room.throw(routing, feathers, caller=self)

    def initialize(self):
        pass

    def _throw(self, pillow, feather, me=None):
        if not me: me = self
        tokens = str(pillow).split('.')
        if len(tokens) > 1 and tokens[1] == "Out":
            self.__room.throw(pillow, feather, me=me)
        else:
            self.__room.throw(pillow, feather, me=me)

    def _uncatch(self, pillow, catcher, passThrower=False):
        ident = pillow, catcher, passThrower

        self.__room.uncatch(*ident)

        try:
            self._catchers.remove(ident)
        except:
            pass

    def _catch(self, pillow, catcher, passThrower=False):
        ident = pillow, catcher, passThrower

        tokens = str(pillow).split('.')
        if pillow == '*' or len(tokens) > 1 and tokens[1] == "In":
            self.__room.catch(*ident)
        else:
            #print "Deprecated catcher", pillow, "in", self.__class__.__name__
            #TODO: REFACTOR PEERS
            self.__room.catch(*ident)

        self._catchers.add(ident)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from uiPeer import *

from wallaby.common.pathHelper import SelectionNotSpecified, PathHelper
from viewer import Viewer
from editor import Editor
from documentChanger import DocumentChanger

class EmbeddedViewer(UIPeer):
    Select = Pillow.In

    LIST = 1
    DICT = 2

    def __init__(self, room, path, delegate, conflictCB=None, isList=True, wrapInList=False, identifier=None, editOnInsert=False):
        UIPeer.__init__(self, room, path)

        self._delegate = delegate
        self._document = None
        self._enabled  = False
        self._conflictCB = conflictCB
        self._isList = isList
        self._wrapInList = wrapInList
        self._identifier = identifier

        self._inserted = None

        if editOnInsert:
            self._catch(DocumentChanger.Out.RowInserted, self._rowInserted)

        self._catch(Viewer.In.Document, self._setDocument)
        self._catch(Viewer.In.Refresh, self._refresh)
        self._catch(Editor.In.Enable, self._enable)
        self._catch(EmbeddedViewer.In.Select, self._select)

        self._catch(Editor.In.Conflict, self._conflict)
        self._resolved = False

    def resolve(self, data=None, **ka):
        if len(ka) > 0:
            self._document.resolve(**ka)
            self._throw(Viewer.In.Refresh, "")
        elif data != None:
            self._document.set(self._path, data)
            self._throw(Viewer.In.Refresh, self._path)
    
        self._resolved = True

    def hasImage(self, name):
        return self._document.hasAttachment(name)

    def getImage(self, name):
        return defer.maybeDeferred(self._document.deferredGetAttachment, name)

    def _select(self, pillow, (path, idx) ):
        if self._identifier != None:
            if not path.startswith(self._identifier):
                return

            path = path.replace(self._identifier + ":", "")

        if path != self._path: return
 
        if self._delegate:
            self._delegate.selectRow(idx)

    def fieldChanged(self, key, field):
        path = self._path + "." + unicode(key)
        path = path.replace('__root__.', '')

        if field == "*":
            self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, path))
        else:
            self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, path + "." + field))

    def dataChanged(self, sel):
        path = self._path + ".*"
        path = path.replace('__root__.', '')

        self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, path))

        self._document.select(self._path + '.*', sel) 
        if self._delegate: self._delegate.selectRow(sel)

    def _enable(self, pillow, isEnabled):
        self._enabled = isEnabled

    def select(self, idx):
        if self._document:
            try:
                currentIdx = PathHelper.getValue(self._document.selection, self._path + '.*._selection') 
                # already selected
                if currentIdx == idx: return 

            except: pass

        from documentChanger import DocumentChanger
        self._throw(DocumentChanger.In.Select, (self._path, idx))

    def _rowInserted(self, pillow, (path, key)):
        if self._path is None or self._delegate is None:
            return

        if Viewer.matchPath(self._path, path, mode=Viewer.OUTER):
            self._inserted = key

    def _refresh(self, pillow, path):
        if self._path is None:
            return
        try:
            # selection changed!
            if path == self._path + ".*":
                try:
                    idx = PathHelper.getValue(self._document.selection, path + '._selection') 
                    if self._delegate:
                        self._delegate.selectRow(idx)
                except: pass

            elif Viewer.matchPath(self._path, path, mode=Viewer.OUTER):
                value = None
                if self._document:
                    if self._isList:
                        value = self._document.get(self._path, asList=True, wrap=self._wrapInList)
                    else:
                        value = self._document.get(self._path, asDict=True)

                if value and self._credentials:
                    credential = self._credentials.getFlat(self._path)
                    if credential == None or 'view' in credential or self._path == '_id':
                        if self._delegate: self._delegate.dataCB(value)
                        if not self._resolved:  
                            if self._document.hasConflicts():
                                d = self._document.loadConflicts()
                                d.addCallback(self._conflictsArrived)
                            else:
                                self._noConflict()

                        if self._delegate and self._inserted != None:
                            self._delegate.editRow(self._inserted)
                else:
                    if self._delegate: self._delegate.dataCB(None)

                if self._document != None and self._document.selection != None:
                    # FIXME: Should we clear the selection here
                    # if "*" in self._path:
                    #     self._document.select(self._path, None)

                    sel = PathHelper.getValue(self._document.selection, self._path + '.*._selection')
                    if sel != None and self._delegate != None: self._delegate.selectRow(sel)

            elif Viewer.matchPath(self._path, path, mode=Viewer.INNER):
                if self._delegate: self._delegate.changedCB()

        except SelectionNotSpecified as e:
            self._delegate.dataCB(None)

        self._inserted = None

    def _conflictsArrived(self, conflicts):
        if conflicts == None: 
            self._noConflict()
            return

        self._resolved = False
        self._conflict(None, conflicts)

    def _noConflict(self):
        if self._conflictCB:
            self._conflictCB(False, None, None)

    def _conflict(self, pillow, feathers): 
        if self._conflictCB: 
            conflicts = [] 

            value = self._document.get(self._path)
 
            if isinstance(feathers, list): 
                for conflict in feathers: 
                    user = conflict.get("wallabyUser")
                    conflicts.append((user, conflict.get(self._path))) 
            else: 
                user = feathers.get("wallabyUser")
                conflictValue = feathers.get(self._path)

                import json
                jsonString = json.dumps(conflictValue)
                jsonString2 = json.dumps(value)

                if jsonString != jsonString2:
                    conflicts.append((user, feathers.get(self._path))) 
         
            if len(conflicts) > 0: 
                self._conflictCB(True, value, conflicts) 
            else: 
                self._conflictCB(False, value, None)

    def _setDocument(self, pillow, document):
        self._document = document
        self._resolved = False
        self._refresh(pillow, self._path)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from uiPeer import *

from wallaby.common.pathHelper import SelectionNotSpecified
from viewer import Viewer

class Editor(UIPeer):

    Enable   = Pillow.InState
    Conflict = Pillow.In

    FieldChanged = Pillow.Out

    def __init__(self, room, registerFieldChangedCallback=None, valueCallback=None, setReadOnlyCallback=None, conflictCallback=None, path=None, selectOnEdit=False, selectCallback=None, raw=False, refreshDoneCallback=None):
        UIPeer.__init__(self, room, path)

        self._valueCallback = valueCallback
        self._setReadOnlyCallback = setReadOnlyCallback
        self._conflictCallback = conflictCallback
        self._selectOnEdit = selectOnEdit
        self._selectCallback = selectCallback
        self._refreshDoneCallback = refreshDoneCallback
        self._document = None
        self._readonly = True
        self._isReadOnly = True
        self._raw = raw

        self._catch(Viewer.In.Document, self._setDocument)
        self._catch(Editor.In.Enable, self._setEnabled)
        self._catch(Editor.In.Conflict, self._conflict)
        self._catch(Viewer.In.Refresh, self._refresh)
        self._catch(Viewer.In.RefreshDone, self._refreshDone)

        self._resolved = False

        if registerFieldChangedCallback:
            registerFieldChangedCallback(self._fieldChanged)

    def _setDocument(self, pillow, feathers):
        self._document = feathers
        self._resolved = False
        self._refresh(None, self._path)

    def _resolve(self, **ka):
        if len(ka) > 0:
            self._document.resolve(**ka)
            self._throw(Viewer.In.Refresh, "")

        self._resolved = True

    def isReadOnly(self):
        return self._isReadOnly

    def _setReadOnly(self, ro):
        if ro != self._isReadOnly and not ro and self._selectOnEdit and self._selectCallback != None:
            self._selectCallback()

        self._isReadOnly = ro

        if self._setReadOnlyCallback:
            self._setReadOnlyCallback(ro)

    def _refreshDone(self, pillow, path):
        if self._refreshDoneCallback:
            self._refreshDoneCallback()

    def _refresh(self, pillow, path):
        if self.destroyed: return

        if path == None or self._readonly:
            self._setReadOnly(True)
            self._noConflict()
            return

        if self._document:
            try:
                if Viewer.matchPath(self._path, path, mode=Viewer.OUTER):
                    self._document.get(self._path)

                    if self._credentials:
                        credential = self._credentials.getFlat(self._path)
                        if credential == None or 'edit' in credential: #Hide whole widget
                            self._setReadOnly(self._readonly)
                            if not self._resolved:
                                if self._document.hasConflicts():
                                    d = self._document.loadConflicts()
                                    d.addCallback(self._conflictsArrived)
                                else:
                                    self._noConflict()

            except SelectionNotSpecified as e:
                self._setReadOnly(True)
        else:
            self._setReadOnly(True)

    def _conflictsArrived(self, conflicts):
        if conflicts == None: 
            self._noConflict()
            return

        self._resolved = False

        self._conflict(None, conflicts)

    def _setEnabled(self, pillow, enabled):
        self._readonly = not enabled
        self._refresh(None, self._path)

    def _noConflict(self):
        if self._conflictCallback:
            self._conflictCallback(False, None, None)

    def _conflict(self, pillow, payload):
        if '*' in self._path: return

        if self._conflictCallback:
            value = self._document.get(self._path)
            conflicts = []

            if isinstance(payload, list):
                for conflict in payload:
                    conflictValue = conflict.get(self._path)
                    user = conflict.get("wallabyUser")
                    if conflictValue != value: conflicts.append((user, conflictValue))
            else:
                conflictValue = payload.get(self._path)
                user = payload.get("wallabyUser")
                if conflictValue != value: conflicts.append((user, conflictValue))
        
            if len(conflicts) > 0:
                self._conflictCallback(True, value, conflicts)
            else:
                self._conflictCallback(False, value, None)

    def changeValue(self, value):
        if self._readonly: return

        if (self._document):
            try:
                if self._document.get(self._path) != value:
                    self._document.set(self._path, value) 
                    self._throwFieldChanged(self._path)
            except SelectionNotSpecified as e:
                self._setReadOnly(True)

    @staticmethod
    def translatePath(doc, path):
        if doc != None and "*" in path:
            try:
                trans = doc.getSelection(path)
                return trans
            except:
                return None
        else:
            return path

    def _throwFieldChanged(self, path):
        self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, path))

    def _fieldChanged(self):
        if self.destroyed: return

        value = None
        if self._valueCallback:
            if self._raw:
                value = self._valueCallback()
            else:
                oldValue = None
                try:
                    oldValue = self._document.get(self._path)
                except:
                    pass

                if oldValue == None or isinstance(oldValue, str) or isinstance(oldValue, unicode):
                    value = unicode(self._valueCallback())
                else:
                    value = self._valueCallback()

                    if isinstance(value, float) or isinstance(oldValue, float):
                        value = float(value)

                    elif isinstance(oldValue, int) or isinstance(value, int):
                        value = int(value) 
                    
                    else:
                        value = unicode(self._valueCallback())

        self.changeValue(value)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

# this is a namespace package
try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
except:
    import pkgutil
    __path__ = pkgutil.extend_path(__path__, __name__)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from uiPeer import *

from wallaby.common.queryDocument import QueryDocument
from documentCache import DocumentCache

class MultiViewer(UIPeer):

    Data = Pillow.InState
    DataChanged = Pillow.In
    Refresh = Pillow.In
    ClearSelection = Pillow.In
    SelectID = Pillow.In

    Select = Pillow.Out
    MultiSelect = Pillow.Out
    Query  = Pillow.OutState
    QueryDocument = Pillow.OutState

    ObjectType = Pillow.Runtime

    def __init__(self, room=None, view=None, viewIdentifier=None, viewArguments=None, dataView=None, delegate=None, autoUpdate=True, orderPath=None, queryDocID=None, **ka):
        UIPeer.__init__(self, room)
        self._delegate    = delegate

        self._view           = view
        self._viewIdentifier = viewIdentifier
        self._viewArguments  = viewArguments
        self._dataView       = dataView
        self._orderPath      = orderPath
        self._queryDocID     = queryDocID

        self._autoUpdate = autoUpdate

        self._queryResult = None
        self._nextQueryResult = None
        # self._initialized = False

        if self._queryDocID != None and len(self._queryDocID) > 0:
            self._catch(DocumentCache.Out.RequestedDocument, self._queryDocArrived)
            self._queryDocument = None
        else:
            self._queryDocument = QueryDocument({
                    'view': self._view,
                    'postfix': '',
                    'identifier': self._viewIdentifier,
                    'dataView': self._dataView,
                    'args': self._viewArguments
            })

        self._catch(MultiViewer.In.Data, self._refreshData)
        self._catch(MultiViewer.In.DataChanged, self._updateData)
        self._catch(MultiViewer.In.Refresh, self._updateQuery)
        self._catch(MultiViewer.In.ClearSelection, self._clearSelection)
        self._catch(MultiViewer.In.SelectID, self._selectID)

    def getImage(self, *args, **ka):
        if not self._queryResult: 
            d = defer.Deferred()
            d.callback(None)
            return d

        return self._queryResult.getImage(*args, **ka)
 

    def sort(self, field, order):
        if not self._queryDocument: return

        if self._queryDocument.get("args") != None:
            if field == '__default__':
                self._queryDocument.set('postfix', '')
            else:
                self._queryDocument.set('postfix', 'By' + field.capitalize())
    
            self._queryDocument.set('args.sort', field)
    
            desc  = self._queryDocument.get('args.descending')
            start = self._queryDocument.get('args.startkey')
            end   = self._queryDocument.get('args.endkey')
            if desc == None: desc = False
    
            if order == 0:
                self._queryDocument.set('args.descending', False)
                if desc and start != None and end != None:
                    self._queryDocument.set('args.startkey', end)
                    self._queryDocument.set('args.endkey', start)
    
            else:
                self._queryDocument.set('args.descending', True)
                if not desc and start != None and end != None:
                    self._queryDocument.set('args.startkey', end)
                    self._queryDocument.set('args.endkey', start)
        else:
            sort = self._queryDocument.get("sort")
            if sort == None or len(sort) == 0: 
                sort = [{field:{"order": None}}]
                self._queryDocument.set("sort", sort)
            elif field not in sort[0]:
                sort[0] = {field:{"order": None}}

            if isinstance(sort[0][field], dict):
                if order == 0:
                    sort[0][field]["order"] = "asc"
                else:
                    sort[0][field]["order"] = "desc"
            else:
                if order == 0:
                    sort[0][field] = "asc"
                else:
                    sort[0][field] = "desc"

        self._updateQuery()

    def initialize(self):
        if self._queryDocID is not None and len(self._queryDocID) > 0:
            self._throw(DocumentCache.In.RequestDocument, self._queryDocID)
        else:
            self._queryDocArrived(None, self._queryDocument)

    def _queryDocArrived(self, pillow, doc):
        if doc == None: return
        if doc != self._queryDocument and doc.documentID != self._queryDocID: return

        self._queryDocument = doc

        if self._viewIdentifier != None:
            viewRoom = self._viewIdentifier.upper()
            self._throw(viewRoom + ':' + MultiViewer.Out.QueryDocument, self._queryDocument)

        if self._autoUpdate:
            self._updateQuery()

    def _clearSelection(self, pillow, feathers):
        self._delegate.clearSelection()

    def _selectID(self, pillow, id):
        if not self._queryResult: return

        for i in range(self.length()):
            if id == self._queryResult.getDocumentID(i):
                self._delegate.selectRow(i)

    def doMultiSelect(self, selection):
        ids = [] 
        for s in selection:
            if s != None and s < self.length():
                ids.append(self._queryResult.getDocumentID(s))

        self._throw(MultiViewer.Out.MultiSelect, ids)

    def doSelect(self, idx, tab):
        if idx == None:
            self._throw(MultiViewer.Out.Select, (None, None, None))
        else:
            self._throw(MultiViewer.Out.Select, (self._queryResult.getDocumentID(idx), tab, idx))

    def _updateQuery(self, pillow=None, feathers=None):
        self._throw(MultiViewer.Out.Query, self._queryDocument)

    def length(self):
        if self._queryResult == None:
            return 0
        else:
            l = self._queryResult.length()
            if l <= 0:
                return 0
            else:
                return l

    def deferredGetValue(self, row, path):
        if self._queryResult == None: return None
        return self._queryResult.deferredGetValue(row, path)

    def reorder(self, *args):
        if self._queryResult == None: return None
        return self._queryResult.reorder(*args)

    def deferredGetDocument(self, row):
        if self._queryResult == None: return None
        return self._queryResult.deferredGetDocument(row)

    def _doRefreshData(self, queryResult):
        if self._delegate != None:
            self._delegate.dataCB(queryResult)

    def _updateData(self, pillow, document):
        if self._delegate != None:
            self._delegate.changedCB(document.documentID)

    def _refresh(self, pillow, path):
        self._refreshData(pillow, self._nextQueryResult)

    def _refreshData(self, pillow, queryResult):
        if queryResult == None:
            self._doRefreshData(None)
            return

        if queryResult.identifier() == None or queryResult.identifier() == self._viewIdentifier:
            initialData = False

            self._nextQueryResult = queryResult

            if self._queryResult == None and queryResult != None:
                initialData = True

            if self._credentials:
                self._queryResult = queryResult
                self._queryResult.setOrderPath(self._orderPath)

                credential = self._credentials.getFlat("VIEW." + unicode(self._viewIdentifier))
                if credential == None or 'view' in credential: #TODO: Hide whole widget
                    self._doRefreshData(queryResult)
                    if initialData and self._delegate:
                        from twisted.internet import reactor
                        reactor.callLater(0.3, self._delegate.initialData)
            else:
                self._doRefreshData(None)





# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

class Draggable(object):
    def dragText(self):
        return None

    def dragURL(self):
        return None

    def dragCustom(self):
        pass



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class PingPongParty(Peer):
    def __init__(self, room):
        Peer.__init__(self, room)

        from pingPong import PingPong
        from ping import Ping
        from pong import Pong

        #Peers
        PingPong(room)

        Ping(room)
        Pong(room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from twisted.internet.protocol import ClientFactory, Protocol
from datetime import *
from twisted.internet import defer
import re

class SDEditChannelProtocol(Protocol):
    def __init__(self):
        self._data = ""
        self._peers = {}
        self._thread = {}
        self._queue = []
        self._room = "UNKNOWN"
        self._connected = False

    def configure(self, peers=None, room=None, pillows=None, translations=None):
        self._peers = peers
        self._room = room
        self._pillows = pillows
        self._translations = translations

    def sendMessage(self, *args):
        if not self._connected:
            self._queue.append(args)
        else:
            self.doSendMessage(*args)

    def doSendMessage(self, sender, pillow, feathers):
        if pillow in self._pillows:
            if (isinstance(feathers, unicode) or isinstance(feathers, str)) and len(feathers) == 32:
                feathers = "..." + feathers[27:31]

            feathers = str(feathers)
            feathers = feathers.replace(':', '=')
            pillowStr = pillow + '(' + feathers + ')'

            if sender in self._translations:
                sender = self._translations[sender]

            if sender not in self._peers: return

            receivers = self._pillows[pillow].keys() 
            if len(receivers) > 1:
                self.transport.write(sender + ':>{' + (','.join(receivers)) + '}.' + pillowStr + '\n')
            elif len(receivers) == 1:
                self.transport.write(sender + ':>' + receivers[0] + '.' + pillowStr + '\n')

            for r in receivers:
                self.transport.write(r + ':stop\n')

    def connectionMade(self):
        self._connected = True
        self.transport.write(self._room+'\n')
        self.transport.write('Room:Actor\n')

        for h in self._peers.keys():
            self.transport.write(h+':'+h+'[a]\n')

        self.transport.write('\n')
        self.transport.write('room:{'+(','.join(self._peers.keys()))+'}.initialize\n')

        while len(self._queue) > 0:
            args = self._queue.pop()
            self.doSendMessage(*args)

    def connectionLost(self, reason):
        pass

    def dataReceived(self, data):
        pass

class SDEditChannelFactory(ClientFactory):
    protocol = SDEditChannelProtocol

    def __init__(self):
        self._config = None
        self._connectedProtocol = None
        self._queue = []

    def configure(self, **args):
        self._config = args

    def sendMessage(self, *args):
        if self._connectedProtocol != None:
            self._connectedProtocol.sendMessage(*args)
        else:
            self._queue.append(args)

    def buildProtocol(self, address):
        self._connectedProtocol = ClientFactory.buildProtocol(self, address)

        if self._config != None: 
            self._connectedProtocol.configure(**self._config)

        while len(self._queue) > 0:
            args = self._queue.pop()
            self._connectedProtocol.sendMessage(*args)

        return self._connectedProtocol



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class CredentialsRouting(Peer):
    from credentialsConsumer import CredentialsConsumer
    from credentials import Credentials

    Routings = [
        (Credentials.Out.Credential, CredentialsConsumer.In.Credential)
    ]

    def __init__(self, room): Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Tab(Peer):
    Select = Pillow.In
    SelectByName = Pillow.In

    def __init__(self, room, delegate, path):
        Peer.__init__(self, room)
        self._delegate = delegate
        self._path = path
        self._catch(Tab.In.Select, self._select)
        self._catch(Tab.In.SelectByName, self._selectByName)

    def _select(self, pillow, path):
        name, idx = path.split('.')
        if self._path == name:
            self._delegate.selectByIndex(int(idx))

    def _selectByName(self, pillow, path):
        base, name = path.split('.')
        if self._path == base:
            self._delegate.selectByName(name)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from document import Document
from wallaby.common.pathHelper import PathHelper

class QueryDocument (Document):
    def __init__(self, data = None):
        Document.__init__(self, "query")
        if data == None: data = {}
        self._data = data

    def get(self, path, **ka):
        return PathHelper.getValue(self._data, path, **ka)

    def set(self, path, value):
        return PathHelper.setValue(self._data, path, value)

    def merge(self, path, value):
        return PathHelper.setValue(self._data, path, value, merge=True)

    def identifier(self):
        return self.get("identifier")



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from viewer import Viewer

class BooleanViewer(Viewer):
    def _refresh(self, pillow, path):
        from wallaby.common.pathHelper import SelectionNotSpecified

        try:
            if Viewer.matchPath(self._path, path, mode=Viewer.OUTER):
                value = False
                if self._document:
                    value = bool(self._document.get(self._path))
                if value and self._credentials:
                    credential = self._credentials.getFlat(self._path)
                    if credential == None or 'view' in credential or self._path == '_id':
                        self._cb(value)
                else:
                    self._cb(False)
        except SelectionNotSpecified as e:
            self._cb(False)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from wallaby.pf.peer.viewer import Viewer

class ActionButtonPeer(Peer):

    Receiving = [
        Viewer.In.Document
    ]

    def __init__(self, room, pillow, feather, tab=None, otherPillows=None, translate=False):
        self.pillow = pillow
        self.feather = feather
        self._otherPillows = otherPillows
        self._translate = translate

        Peer.__init__(self, room)

        if tab != 'None':
            self.tab = tab
        else:
            self.tab = None

        self._doc = None

        if self._translate:
            self._catch(Viewer.In.Document, self._docArrived)

    def _docArrived(self, pillow, doc):
        self._doc = doc

    def dynamicPillows(self):
        if isinstance(self.pillow, (unicode, str)):
            return [self.pillow]
        else:
            return []

    def __translate(self, feathers):
        if feathers != None:
            import re
            if isinstance(feathers, (str, unicode)) and '%__path__:' in feathers:
                if self._doc is None: return None

                match = re.match('(.*?)%__path__:(.*?)%(.*)', feathers)
                return match.group(1) + self._doc.get(match.group(2)) + match.group(3)
            if isinstance(feathers, dict):
                import copy
                feathers = copy.deepcopy(feathers)
                for k, v in feathers.items():
                    if '%__path__:' in v:
                        if self._doc is None: return None

                        match = re.match('(.*?)%__path__:(.*?)%(.*)', v)
                        feathers[k] = match.group(1) + self._doc.get(match.group(2)) + match.group(3)

                return feathers
            
        return feathers 

    def buttonClicked(self, feather=None):
        if self.feather is not None: 
            if not isinstance(self.feather, (unicode, str)) or len(self.feather) > 0: feather = self.feather 

        if self._translate: feather = self.__translate(feather)
            
        self._throw(self.pillow, feather)

        if self._otherPillows:
            for pillow in self._otherPillows:
                _feather = None

                if ';' in pillow:
                    pillow, _feather = pillow.split(';')

                if _feather is None and feather is not None: _feather = feather

                if self._translate: _feather = self.__translate(_feather)

                self._throw(pillow, _feather)

        if self.tab:
            room = self._roomName
            tab = self.tab
            if ":" in tab:
                room, tab = tab.split(":")

            from tab import Tab
            self._throw(room + ":" + Tab.In.Select, tab)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class ValueQueryResultDocument(Peer):
    from valueQueryResult import ValueQueryResult
    from documentCache import DocumentCache
    from database import Database

    Routings = [
        (DocumentCache.Out.RequestedDocument, ValueQueryResult.In.RequestedDocument),
        (ValueQueryResult.Out.RequestDocument, DocumentCache.In.RequestDocument),
        (ValueQueryResult.Out.SaveDocument, Database.In.SaveDocument),
    ]

    def __init__(self, room):
        Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from multiViewer import MultiViewer
from viewer import Viewer
from editor import Editor

class MemDoc(Peer):

    Routings = [
        (Editor.Out.FieldChanged, Viewer.In.Refresh)
    ]

    def __init__(self, room):
        Peer.__init__(self, room)
       
    def initialize(self):
        from wallaby.common.document import Document
        self._throw(Viewer.In.Document, Document())
        self._throw(Editor.In.Enable, True)






# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class LoadDocument(Peer):

    def __init__(self, srcRoom, dstRoom=None, path=None, pillow='EditDocument.In.Load'):
        from viewer import Viewer

        Peer.__init__(self, srcRoom)
        self._dstRoom = dstRoom
        self._pillow = pillow

        self._peer = Viewer(srcRoom, self._loadDocument, path)

    def destroy(self, remove=False):
        Peer.destroy(self, remove)
        if self._peer: self._peer.destroy(remove)

    def _loadDocument(self, value):
        self._throw(self._dstRoom + ':' + self._pillow, value)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from wallaby.common.queryResult import QueryResult
from functools import partial

import math, re

class ListOfDeferreds(list):
    pass

class ValueQueryResult(Peer, QueryResult):
    
    RequestedDocument = Pillow.In
    RequestDocument = Pillow.Out
    SaveDocument = Pillow.Out

    SERIAL = 1

    def __init__(self, source, query, values, room, count=None):
        Peer.__init__(self, room)
        QueryResult.__init__(self, source, query)

        from documentChanger import DocumentChanger
        DocumentChanger.reregisterToken(room + "NextOrder", self.nextOrderToken)

        self._values = values

        from uuid import uuid4
        self._ident  = uuid4().hex
        self._count  = count
        self._limit  = -1
        self._serial = unicode(ValueQueryResult.SERIAL)
        self._orderPath = None
        ValueQueryResult.SERIAL += 1

        if query != None:
            limit = query.get('args.limit')
            if limit != None:
                self._limit = int(limit)

        self._catch(ValueQueryResult.In.RequestedDocument, self._requestedDocument)
        self._defers = {}
        self._deferredDocs = {}

    def setOrderPath(self, path):
        self._orderPath = path

    def _loadPartial(self, idx, cb):
        if self._limit == -1:
            d = defer.Deferred()
            d.callback(None)
            return d

        l = len(self._values)

        if idx < l and isinstance(self._values[idx], ListOfDeferreds):
            d = defer.Deferred()
            self._values[idx].append( (d, cb) )
            return d

        # fill all skipped values with None (could be loaded later)
        for i in range(l-1,idx):
            self._values.append(None)

        l = len(self._values)

        # fill with deferred list for later requests 
        for i in range(idx, idx+self._limit):
            if i >= self._count: break
            if i >= l: 
                self._values.append(ListOfDeferreds())
            else:
                if not isinstance(self._values[i], ListOfDeferreds):
                  self._values[i] = ListOfDeferreds()

        d = defer.Deferred()
        self._values[idx].append( (d, cb) )

        from twisted.internet import reactor
        reactor.callLater(0, self.__loadPartial, idx)
        return d


    @defer.inlineCallbacks
    def __loadPartial(self, idx):
        # make copy to support concurrent requests!
        import copy
        query = copy.deepcopy(self.query)
        query.set('args.skip', idx)

        values, cnt = yield self._source.updateQuery(query)
        if cnt >= 0:
            self._count = cnt

        l = len(self._values)

        callList = ListOfDeferreds()

        # fill values to value cache and call waiting deferreds 
        j = 0
        for i in range(idx, idx+self._limit):
            if j >= len(values): break

            if isinstance(self._values[i], ListOfDeferreds):
                for deferredAndCB in self._values[i]:
                    callList.append(deferredAndCB)

            self._values[i] = values[j]
            j += 1

        for d, cb in callList:
            d2 = cb()
            d2.addCallback(d.callback)

    def getImage(self, row, path, documentID=None):
        if documentID == None:
            documentID = self.getDocumentID(row)

        if isinstance(path, (list, tuple)):
            path = path[0]

        if documentID not in self._deferredDocs:
            self._deferredDocs[documentID] = defer.Deferred()
            self._throw(ValueQueryResult.Out.RequestDocument, documentID)

        d = defer.Deferred()
        self._deferredDocs[documentID].addCallback(partial(self._getImage, path, d))
        return d

    @defer.inlineCallbacks
    def _getImage(self, path, d, doc):
        from wallaby.qt_combat import QtGui, QtCore

        if isinstance(doc, QtGui.QPixmap):
            d.callback(doc)
            defer.returnValue(doc)
            return

        p = QtGui.QPixmap()

        if doc == None:
            d.callback(p)
            return

        data = yield defer.maybeDeferred(doc.deferredGetAttachment, path)

        if data != None:
            img = QtGui.QImage.fromData(data).scaled(128, 128, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation)
            if img != None:
                p.convertFromImage(img)
        d.callback(p)
        defer.returnValue(p)

    # get value from query result. Load more rows if neccessary 
    def deferredGetValue(self, index, path):
        from twisted.internet import reactor

        if index < 0 or index >= len(self._values) or self._values[index] == None or isinstance(self._values[index], ListOfDeferreds):
            if index < 0 or self._count == None or not self._source or index >= self._count:
                d = defer.Deferred()
                reactor.callLater(0, d.callback, None)
                return d
            else:
                return self._loadPartial(index, partial(self._deferredGetRowValue, index, path))

        return self._deferredGetRowValue(index, path)

    def _deferredGetRowValue(self, index, path):
        documentID = self.getDocumentID(index)

        from wallaby.common.document import Document
        doc = Document(documentID, data=self._values[index])

        return self.deferredGetDocValue(path, documentID, document=doc)

    # get corresponding document for row. Load more rows if neccessary
    def deferredGetDocument(self, index):
        if index < 0 or index >= len(self._values) or self._values[index] == None or isinstance(self._values[index], ListOfDeferreds):
            if index < 0 or self._count == None or not self._source or index >= self._count:
                d = defer.Deferred()
                reactor.callLater(0, d.callback, None)
                return d
            else:
                return self._loadPartial(index, partial(self._deferredGetDocument, index))

        return self._deferredGetDocument(index)

    def reorder(self, row, i1, i2):
        if self._orderPath is None:
            FX.crit("No valid order path")
            return

        d1 = self.deferredGetValue(i1, "key.0")
        d2 = self.deferredGetValue(i2, "key.0")
        d3 = self.deferredGetDocument(row)

        list = defer.DeferredList([d1, d2, d3])
        list.addCallback(partial(self._deferredReorderRows, row, i1, i2))

    @defer.inlineCallbacks
    def _deferredReorderRows(self, row, i1, i2, ((valid1, key1), (valid2, key2), (valid3, doc))):
        if not valid3: 
            FX.crit("Error retrieving document")
            return

        if not valid1 and not valid2:
            FX.crit("No valid orders")
            return

        if valid1 and key1 != None and valid2 and key2 != None:
            if abs(key1 - key2) < 1e-6:
                FX.warn("min-delta reached! Reorder all items")
                doc.set(self._orderPath, (i1 + i2)/2.0)
                docs = [doc]

                for i in range(0, self.length()):
                    if i == row: continue
                    d = yield self.deferredGetDocument(i)
                    d.set(self._orderPath, i)
                    docs.append(d)

                self._throw(ValueQueryResult.Out.SaveDocument, (docs, None))
                return

            doc.set(self._orderPath, (key1 + key2)/2.0)
        elif valid1 and key1 != None:
            doc.set(self._orderPath, math.ceil(key1 + 1.0))
        else: 
            doc.set(self._orderPath, math.floor(key2 - 1.0))

        self._throw(ValueQueryResult.Out.SaveDocument, (doc, None))

    def _deferredGetDocument(self, index):
        documentID = self.getDocumentID(index)
        if documentID not in self._deferredDocs:
            self._deferredDocs[documentID] = defer.Deferred()
            self._throw(ValueQueryResult.Out.RequestDocument, documentID)

        return self._deferredDocs[documentID]

    def deferredGetDocValue(self, path, documentID, document=None):
        if path == None:
            d = defer.Deferred()
            d.callback(None)
            return

        if documentID not in self._defers:
            self._defers[documentID] = {}

        if '(' in path or document==None:
            if path not in self._defers[documentID]:
                self._defers[documentID][path] = defer.Deferred()

                if document != None:
                    from twisted.internet import reactor 
                    reactor.callLater(0, self._requestedDocument, None, document)
                else:
                    self._throw(ValueQueryResult.Out.RequestDocument, documentID)

            return self._defers[documentID][path]
        # Performance boost for existing values 
        else:
            d = defer.Deferred()
            try:
                if '|' in path:
                    pathes = path.split('|')
                    value, i = None, 0
                    while value == None and i < len(pathes):
                        path = pathes[i]
                        value = document.get(path)  
                        i += 1

                    if value == None:
                        d.callback("-")
                    else:
                        d.callback(value)
                else:
                    d.callback(document.get(path))
            except:
                d.callback('-')

            return d

    def _chainCallback(self, d, v): 
        d.callback(v)
        return v

    def _requestedDocument(self, pillow, document):
        from twisted.internet import reactor
        documentID = document.documentID

        if documentID in self._deferredDocs:
            d = self._deferredDocs[documentID]
            del self._deferredDocs[documentID]
            reactor.callLater(0, d.callback, document)

        if documentID in self._defers:
            for path, d in self._defers[documentID].items():
                if '|' in path:
                    pathes = path.split('|')
                else:
                    pathes = [path]

                value, i = None, 0
                while value == None and i < len(pathes):
                    path = pathes[i]

                    if '(' in path:
                        newpath = re.sub(r'\([^)]+\)\.', '', path, 1)
                        m = re.match(r'\(([^)]+)\)\.', path)
                        docID = document.get(m.group(1))

                        if docID != None:
                            newDeferred = self.deferredGetDocValue(newpath, docID)
                            newDeferred.addCallback(partial(self._chainCallback, d)) 
                        else:
                            reactor.callLater(0, d.callback, None)

                        del self._defers[documentID]
                        return
                    else:
                        if '_source' in document._data and "_source" not in path:
                            value = document.get('_source.' + path)
                        else:
                            value = document.get(path)

                    i += 1

                reactor.callLater(0, d.callback, value)

            del self._defers[documentID]

    def getDocumentID(self, index):
        if self._values == None or index >= len(self._values):
            # Dummy documentID for table model expansion
            return self._serial + "_" + unicode(index)

        try:
            if 'value' in self._values[index] and isinstance(self._values[index]['value'], dict) and '_id' in self._values[index]['value']:
                return self._values[index]['value']['_id']
        except Exception as e:
            print "Exception:", self._values[index], self.query._data, e

        if '_id' in self._values[index]:
            return self._values[index]['_id']

        if 'id' in self._values[index]:
            return self._values[index]['id']

        return self._ident + str(index)

    @defer.inlineCallbacks
    def nextOrderToken(self, doc):
        if self._values != None and self._orderPath != None:
            order = yield self.deferredGetValue(self.length()-1, "key.0")
            defer.returnValue(order+1)
            return

        defer.returnValue(self.length())

    def length(self):
        if self._count != None:
            return self._count
        else:
            return len(self._values)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from viewer import Viewer

class Observer(Peer):
    from editor import Editor
    from documentChanger import DocumentChanger

    Sending = [
        Viewer.In.Document
    ]

    Routings = [
        (Editor.Out.FieldChanged, Viewer.In.Refresh),
        (DocumentChanger.Out.SelectionChanged, Viewer.In.Refresh)
    ]

    def __init__(self, room):
        Peer.__init__(self, room)

    def initialize(self):
        from wallaby.pf.room import House
        observer = House.observer()

        from wallaby.common.document import Document
        doc = Document()

        peerNames = sorted(observer.allPeers().keys())

        peers = []

        for peer in peerNames:
            ibp = observer.inBoundPillows(peer)
            obp = observer.outBoundPillows(peer)

            peers.append({
                "name": peer,
                "inBound": ibp,
                "outBound": obp
            })

        doc.set("peers", peers)
        self._throw(Viewer.In.Document, doc)
        
        # Wildcard credentials
        from credentials import Credentials
        self._throw(Credentials.Out.Credential, Document())



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

import os, string, re
import wallaby.FX as FX

class Observer(object):
    def __init__(self):
        self._inBoundPillows = {}
        self._inBoundPillowsFQ = {}
        self._outBoundPillows = {}
        self._outBoundPillowsFQ = {}
        self._inBoundPillowsPerPeer = {}
        self._outBoundPillowsPerPeer = {}
        self._objectTypes = {}
        self._extendedBy = {}
        self._bases = {}
        self._peers = {}
        self._rooms = {}
        self._fqn = {}

    def peerClass(self, name):
        if name in self._peers:
            return self._peers[name]
        else:
            return None

    def peer(self, name, *args, **ka):
        if name in self._peers:
            return self._peers[name](*args, **ka)
        else:
            return None

    def allPeers(self):
        return self._peers

    def allInBoundPillows(self):
        return self._inBoundPillowsFQ

    def allOutBoundPillows(self):
        return self._outBoundPillowsFQ

    def peers(self, pillow, lst):
        if not pillow in lst:
            return []
        else:
            peers = lst[pillow]
            allPeers = {}
            self._appendPeers(peers, allPeers)
            return allPeers.keys()

    def objectType(self, peer):
        if not peer in self._objectTypes:
            return None
        else:
            return self._objectTypes[peer]

    def _appendPeers(self, _from, _to):
        for peer in _from:
            _to[peer] = True 
            if peer in self._extendedBy:
                self._appendPeers(self._extendedBy[peer], _to)

    def pillows(self, peer, lst):
        if peer in lst:
            pillows = lst[peer]
        else:
            pillows = []

        allPillows = {}
        self._appendPillows(peer, lst, pillows, allPillows)
        return allPillows.keys()

    def _appendPillows(self, peer, lst, _from, _to):
        for pillow in _from:
            _to[pillow] = True 

        if peer in self._bases:
            for h in self._bases[peer]:
                if h in lst:
                    pillows = lst[h]
                else:
                    pillows = []

                self._appendPillows(h, lst, pillows, _to)

    def outBoundPillows(self, peer=None):
        if peer==None:
            return self._outBoundPillows.keys()
        else:
            if '.' not in peer and peer in self._fqn: 
                peer = self._fqn[peer]
               
            return self.pillows(peer, self._outBoundPillowsPerPeer)

    def inBoundPillows(self, peer=None):
        if peer==None:
            return self._inBoundPillows.keys()
        else:
            if '.' not in peer and peer in self._fqn: 
                peer = self._fqn[peer]

            return self.pillows(peer, self._inBoundPillowsPerPeer)

    def outBoundPeers(self, pillow=None, fq=False):
        if pillow==None:
            return self._outBoundPillowsPerPeer.keys()
        else:
            pillow = pillow.replace('!', '')
            if fq:
                return self.peers(pillow, self._outBoundPillowsFQ)
            else:
                return self.peers(pillow, self._outBoundPillows)

    def inBoundPeers(self, pillow=None, fq=False):
        if pillow==None:
            return self._inBoundPillowsPerPeer.keys()
        else:
            pillow = pillow.replace('!', '')
            if fq:
                return self.peers(pillow, self._inBoundPillowsFQ)
            else:
                return self.peers(pillow, self._inBoundPillows)

    def scan(self, peers=None):
        fqPeers = []

        if peers is None:
            import wallaby.pf.peer as peers
            from twisted.plugin import getCache
            peers = getCache(peers)
            root = "wallaby.pf.peer"

            for peer in peers:
                fqPeers.append(root + "." + peer)
        else:
            fqPeers = peers

        for fqPeer in fqPeers:
            
            mod = FX.imp(fqPeer)
            root, basename = fqPeer.rsplit('.', 1)

            if mod == None: continue

            cls = string.upper(basename[0]) + basename[1:]                       

            if not cls in mod.__dict__: continue

            fqn = root + '.' + basename + '.' + cls

            self._fqn[cls] = fqn

            if re.search(r"\.room", root) != None:
                self._rooms[cls] = mod.__dict__[cls]
            else:
                self._peers[cls] = mod.__dict__[cls]

            self._objectTypes[fqn] = mod.__dict__[cls].ObjectType

            for base in mod.__dict__[cls].__bases__:
                fqBase = base.__module__ + '.' + base.__name__
                if not fqBase in self._extendedBy:
                    self._extendedBy[fqBase] = set()

                self._extendedBy[fqBase].add(fqn)

                if not fqn in self._bases:
                    self._bases[fqn] = set()

                self._bases[fqn].add(fqBase)

            if 'Receiving' in mod.__dict__[cls].__dict__:
                for fqa in mod.__dict__[cls].Receiving:
                    pillow = None
                    try:
                        pillow = fqa.split('.')[2]
                    except:
                        print "Split error:", fqa

                    pillow = pillow.replace('!', '')
                    fqa = fqa.replace('!', '')

                    if not pillow in self._inBoundPillows: self._inBoundPillows[pillow] = []
                    if not fqa in self._inBoundPillowsFQ: self._inBoundPillowsFQ[fqa] = []
                    if not fqn in self._inBoundPillowsPerPeer: self._inBoundPillowsPerPeer[fqn] = []
                    self._inBoundPillows[pillow].append(fqn)
                    self._inBoundPillowsFQ[fqa].append(fqn)
                    self._inBoundPillowsPerPeer[fqn].append(fqa)


            for pillow in mod.__dict__[cls].InPillows:
                pillow = pillow.replace('!', '')

                fqa = cls + '.In.' + pillow
                fqa = fqa.replace('!', '')

                if not pillow in self._inBoundPillows: self._inBoundPillows[pillow] = []
                if not fqa in self._inBoundPillowsFQ: self._inBoundPillowsFQ[fqa] = []
                if not fqn in self._inBoundPillowsPerPeer: self._inBoundPillowsPerPeer[fqn] = []
                self._inBoundPillows[pillow].append(fqn)
                self._inBoundPillowsFQ[fqa].append(fqn)
                self._inBoundPillowsPerPeer[fqn].append(fqa)

            if 'Sending' in mod.__dict__[cls].__dict__:
                for fqa in mod.__dict__[cls].Sending:
                    pillow = None
                    try:
                        pillow = fqa.split('.')[2]
                    except:
                        print "Split error:", fqa

                    fqa = fqa.replace('!', '')
                    pillow = pillow.replace('!', '')

                    if not pillow in self._outBoundPillows: self._outBoundPillows[pillow] = []
                    if not fqa in self._outBoundPillowsFQ: self._outBoundPillowsFQ[fqa] = []
                    if not fqn in self._outBoundPillowsPerPeer: self._outBoundPillowsPerPeer[fqn] = []
                    self._outBoundPillows[pillow].append(fqn)
                    self._outBoundPillowsFQ[fqa].append(fqn)
                    self._outBoundPillowsPerPeer[fqn].append(fqa)

            for pillow in mod.__dict__[cls].OutPillows:
                pillow = pillow.replace('!', '')

                fqa = cls + '.Out.' + pillow
                fqa = fqa.replace('!', '')
                if not pillow in self._outBoundPillows: self._outBoundPillows[pillow] = []
                if not fqa in self._outBoundPillowsFQ: self._outBoundPillowsFQ[fqa] = []
                if not fqn in self._outBoundPillowsPerPeer: self._outBoundPillowsPerPeer[fqn] = []
                self._outBoundPillows[pillow].append(fqn)
                self._outBoundPillowsFQ[fqa].append(fqn)
                self._outBoundPillowsPerPeer[fqn].append(fqa)





# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from wallaby.common.queryResult import QueryResult
from twisted.internet import defer
from peer import *
from functools import *

class DocumentIDQueryResult(Peer, QueryResult):
    RequestedDocument = Pillow.In
    RequestDocument = Pillow.Out

    ObjectType = Pillow.Runtime

    def __init__(self, source, query, ids, room, count=None):
        Peer.__init__(self, room)
        QueryResult.__init__(self, source, query)

        self._ids = ids
        self._count = count

        self._catch(DocumentIDQueryResult.In.RequestedDocument, self._requestedDocument)
        self._defers = {}
        self._deferredDocs = {}
        self._orderPath = None

    def setOrderPath(self, path):
        self._orderPath = path

    def deferredGetValue(self, index, path):
        from twisted.internet import reactor

        if index < 0 or index >= self.length():
            d = defer.Deferred()
            reactor.callLater(0, d.callback, None)
            return d
        else:
            documentID = self.getDocumentID(index)

        return self.deferredGetDocValue(documentID, path)

    def getImage(self, row, path, documentID=None):
        if documentID == None:
            documentID = self.getDocumentID(row)

        if isinstance(path, (list, tuple)):
            path = path[0]

        if documentID not in self._deferredDocs:
            self._deferredDocs[documentID] = defer.Deferred()
            self._throw(DocumentIDQueryResult.Out.RequestDocument, documentID)

        d = defer.Deferred()
        self._deferredDocs[documentID].addCallback(partial(self._getImage, path, d))
        return d

    @defer.inlineCallbacks
    def _getImage(self, path, d, doc):
        from wallaby.qt_combat import QtGui, QtCore

        if isinstance(doc, QtGui.QPixmap):
            d.callback(doc)
            defer.returnValue(doc)
            return

        p = QtGui.QPixmap()

        if doc == None:
            d.callback(p)
            return

        data = yield defer.maybeDeferred(doc.deferredGetAttachment, path)

        if data != None:
            img = QtGui.QImage.fromData(data).scaled(128, 128, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation)
            if img != None:
                p.convertFromImage(img)
        d.callback(p)
        defer.returnValue(p)



    def deferredGetDocument(self, index):
        from twisted.internet import reactor

        if index >= self.length() or index < 0:
            d = defer.Deferred()
            reactor.callLater(0, d.callback, None)
            return d

        documentID = self.getDocumentID(index)
        if documentID not in self._deferredDocs[documentID]:
            self._deferredDocs[documentID] = defer.Deferred()
            self._throw(DocumentIDQueryResult.Out.RequestDocument, documentID)

        return self._deferredDocs[documentID]

    def deferredGetDocValue(self, documentID, path):
            if documentID not in self._defers:
                self._defers[documentID] = {}

            if path not in self._defers[documentID]:
                self._throw(DocumentIDQueryResult.Out.RequestDocument, documentID)
                self._defers[documentID][path] = defer.Deferred()

            return self._defers[documentID][path]

    def _chainCallback(self, d, v):
        d.callback(v)
        return v

    def _requestedDocument(self, pillow, document):
        from twisted.internet import reactor
        documentID = document.documentID

        if documentID in self._deferredDocs:
            d = self._deferredDocs[documentID]
            del self._deferredDocs[documentID]
            reactor.callLater(0, d.callback, document)

        if documentID in self._defers:
            for path, d in self._defers[documentID].iteritems():
                if '(' in path:
                    import re
                    newpath = re.sub(r'\([^)]+\)\.', '', path, 1)
                    m = re.match(r'\(([^)]+)\)\.', path)
                    docID = document.get(m.group(1))

                    if docID != None:
                        newDeferred = self.deferredGetDocValue(docID, newpath)

                        from functools import partial
                        newDeferred.addCallback(partial(self._chainCallback, d))
                    else:
                        reactor.callLater(0, d.callback, None)
                else:
                    reactor.callLater(0, d.callback, document.get(path))

            del self._defers[documentID]

    def getDocumentID(self, index):
        if index >= self.length():
            return None
        return self._ids[index]

    def length(self):
        return len(self._ids)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class SMPeer(Peer):
    def __init__(self, room, statePillow, states=[], initState="Start"):
        Peer.__init__(self, room)
        self._routings = set()
        self._transitions = set()
        self._callbacks = set()
        self._states = {} #Name->State
        self._statePillow = statePillow
        self._initState = initState

        self._state = None

        # if len(states) > 0:
        #     self._state = states[0]
        # else:
        #     pass #TODO: Throw exception

        for state in states:
            self._states[state._name] = state
            state._setStateMachine(self)

        for pillow in self._callbacks:
            self._catch(pillow, self._callback)

        for pillow in self._routings:
            self._catch(pillow, self._routePillow)

        for pillow in self._transitions:
            self._catch(pillow, self._transitionState)

    def initialize(self):
        self.switchState(self._initState)

    def addRoutings(self, pillows):
        self._routings = self._routings.union(set(pillows))

    def addTransitions(self, pillows):
        self._transitions = self._transitions.union(set(pillows))

    def addCallbacks(self, pillows):
        self._callbacks = self._callbacks.union(set(pillows))

    def _routePillow(self, *args):
        self._state._routePillow(*args)

    def _transitionState(self, *args):
        self._state._transitionState(*args)

    def _callback(self, *args):
        self._state._callback(*args)

    def switchState(self, stateName):
        # Already in correct state
        if self._state != None and self._state._name == stateName: return

        # print "Switch to state", stateName, "in context", self._roomName
        if stateName in self._states:
            self._state = self._states[stateName]
            self._throw(self._statePillow, stateName)
            self._state._stateSwitched()

class State:
    def __init__(self, name=None):
        if name:
            self._name = name
        else:
            self._name = self.__class__.__name__

        self._stateMachine = None
        self._routings = {}
        self._transitions = {}
        self._callbacks = {}
        self._localCallbacks = {}

    def _stateSwitched(self):
        pass

    def _addRouting(self, sourcePillow, destinationPillow):
        if not sourcePillow in self._routings:
            self._routings[sourcePillow] = set()
        self._routings[sourcePillow].add(destinationPillow)

    def _setTransition(self, pillow, destinationState):
        self._transitions[pillow] = destinationState

    def _catch(self, pillow, callback):

        if not pillow in self._callbacks:
            self._callbacks[pillow] = set()

        self._callbacks[pillow].add(callback)

        if ':' in str(pillow): 
            room, pillow = pillow.split(':')

        if not pillow in self._localCallbacks:
            self._localCallbacks[pillow] = set()

        self._localCallbacks[pillow].add(callback)

    def sm(self):
        return self._stateMachine

    def _setStateMachine(self, stateMachine):
        self._stateMachine = stateMachine
        self._stateMachine.addRoutings(self._routings.keys())
        self._stateMachine.addTransitions(self._transitions.keys())
        self._stateMachine.addCallbacks(self._callbacks.keys())

    def _throw(self, pillow, feathers):
        self._stateMachine._throw(pillow, feathers, self)

    def _switchState(self, state):
        self._stateMachine.switchState(state)

    def _routePillow(self, pillow, feathers):
        if pillow in self._routings:
            for routing in self._routings[pillow]:
                self._throw(routing, feathers)

    def _transitionState(self, pillow, feathers):
        if pillow in self._transitions:
            self._switchState(self._transitions[pillow])

    def _callback(self, pillow, feathers):
        if pillow in self._localCallbacks:
            for callback in self._localCallbacks[pillow]:
                callback(pillow, feathers)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Configuration(Peer):
    Configuration = Pillow.In

    def __init__(self, cb):
        Peer.__init__(self, '__CONFIG__')
        self._cb = cb

        self._catch(Configuration.In.Configuration, self._config)

    def _config(self, pillow, document):
        if self._cb != None:
            self._cb(document)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from viewer import Viewer
from editor import Editor

class SearchDocument(Peer):

    #Receiving pillows:
    Search = Pillow.In

    Sending = [
        Viewer.In.Document
    ]

    Routings = [
        (Editor.Out.FieldChanged, Viewer.In.Refresh)
    ]

    def __init__(self, searchRoom, resultRoom=None, document=None, identifier=None, appendWildcard=False):
        Peer.__init__(self, searchRoom)

        if not resultRoom:
            resultRoom = searchRoom
        self._resultRoom = resultRoom

        self._identifier = identifier
        self._appendWildcard = appendWildcard

        if document:
            self._document = document
        else:
            from wallaby.common.queryDocument import QueryDocument
            self._document = QueryDocument()

        self._document.set('identifier', self._identifier)
        self._document.set('query', '*')

        self._catch(SearchDocument.In.Search, self._search)
    
    
    def initialize(self):
        # set search field editable
        self._throw(Viewer.In.Document, self._document)
        self._throw(Editor.In.Enable, True)

    def _search(self, pillow, feathers):
        query = self._document.get('query')
        if self._appendWildcard and (len(query) == 0 or query[-1] != "*"):
            self._document.set('query', query + "*")

        from abstractQuery import AbstractQuery
        self._throw(self._resultRoom+":"+AbstractQuery.In.Query, self._document)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

try:
    from collections import OrderedDict
except:
    class OrderedDict(object):
        def __init__(self):
            print "Error: OrderedDict not supported!!"

from operator import itemgetter
from twisted.internet import task, defer
import wallaby.FX as FX
import inspect
from functools import partial
from wallaby.pf.pillow import Pillow
from wallaby.observer import *
import sys, copy, os.path

from wallaby.common.sets import *

class House(object):
    rooms = {}
    _observer = None

    @staticmethod
    def get(room):
        if room in (None, ""): room = "None"

        if not room in House.rooms:
            if FX.appModule:
                app = FX.appModule
                if room == "__CONFIG__": app = "wallaby.apps.inspector"

                try:
                    from twisted.plugin import getCache
                    pkg = __import__(app + '.rooms', globals(), locals(), ["*"], 0)
                    modules = getCache(pkg)

                    if room.lower() in modules:
                        mod = FX.imp(app + '.rooms.' + room.lower())
                        if mod:
                            cls = room.capitalize()

                            if cls in mod.__dict__:
                                ctx = House.rooms[room] = mod.__dict__[cls](room)
                                return ctx
                except:
                    pass
                        
            # No room template found. Create generic room
            House.rooms[room] = Room(room)

        return House.rooms[room]

    @staticmethod
    def destroyRoom(room):
        if room in House.rooms:
            House.rooms[room].destroyPeers()
            del House.rooms[room]

    @staticmethod
    def pillowRooms(room, pillow, feathers):    
        rooms = OrderedSet()
        catchers = House.get(room).catchers(pillow)
        if catchers != None and len(catchers) > 0:
            for c in catchers:
                rooms |= c.rooms(pillow, feathers)

        return rooms

    @staticmethod
    def catch(room, pillow, catcher, passThrower=False, single=False):
        room = House.get(room)
        room.catch(pillow, catcher, passThrower, single)

    @staticmethod
    def uncatch(room, pillow, catcher, passThrower=False):
        room = House.get(room)
        room.uncatch(pillow, catcher, passThrower)

    @staticmethod
    def throw(pillow, feathers="", **ka):
        parts = pillow.split(':')
        House.get(parts[0]).throw(parts[1], feathers, **ka)

    @staticmethod
    def observer(basePath=""):
        if House._observer == None:
            House._observer = Observer()

            app = FX.appModule
            try:
                from twisted.plugin import getCache
                pkg = __import__(app + '.peer', globals(), locals(), ["*"], 0)
                modules = getCache(pkg)

                fqPeers = []

                for m in modules:
                    fqPeers.append(app + '.peer.' + m)

                House._observer.scan(fqPeers)
            except:
                pass

            House._observer.scan()

        return House._observer

    @classmethod
    def initializeAll(cls, onlyNew=False):
        for r in House.rooms.values(): 
            if not r.initialized  or not onlyNew:
                r.initialize()

class Room(object):
    __metaclass__ = Pillow

    NoHandlerFound = Pillow.Out
    Ignore = Pillow.In
    Suggests = Pillow.OutState

    Suggest = Pillow.In
    Initialize = Pillow.In

    # registerType = OrderedDict
    # registerType = set
    registerType = list

    def __init__(self, name):
        self.initialized = False

        self._name = name
        self._pillows = {}
        self._singlePillows = {}
        self._pillowGuard = set([self])
        self._inBoundPillows = set()
        self._outBoundPillows = set()
        self._statefullPillows = {}

        self._statefullQueue = {}
        self._normalQueue    = []

        self._dequeueCall    = None

        self._lastConfig = None

        if 'Receiving' in self.__class__.__dict__: 
            self._inBoundPillows |= set(self.__class__.__dict__['Receiving'])

        if 'Sending' in self.__class__.__dict__: 
            self._outBoundPillows |= set(self.__class__.__dict__['Sending'])

        self._allPeers = OrderedSet()
        self._peers = []
        self._loadedPeers = []
        self._dynamicPeers = []

        if 'Peers' in self.__class__.__dict__: 
            self._peers = self.__class__.__dict__['Peers']

        # self._config = {'Peers': []}
        self._unclean = False
        self.catch(Room.In.Ignore, self._ignore)

        self._configViewer = None
        self._configDoc = None

        from wallaby.pf.peer.viewer import Viewer

        # None room is not supported
        if self._name == "None": return

        if self._name != "__CONFIG__":
            House.get("__CONFIG__").catch(Viewer.In.RefreshDone, self.configUpdated)
            House.get("__CONFIG__").catch(Viewer.In.Refresh, self._configChanged)
            House.get("__CONFIG__").catch(Viewer.In.Document, self._configDocChanged)
        else:
            self.catch(Viewer.In.RefreshDone, self.configUpdated)
            self.catch(Viewer.In.Refresh, self._configChanged)
            self.catch(Viewer.In.Document, self._configDocChanged)

        self.catch(Room.In.Suggest, self.throwSuggests)
        self.catch(Room.In.Initialize, self.configUpdated)

        # if self._name not in ("__CONFIG__", "__CREDENTIALS__"):
        #     self._configViewer = Viewer("__CONFIG__", self._configChanged, "rooms." + self._name, raw=True, refreshDoneCallback=self.configUpdated, ignoreCredentials=True)
    def _configDocChanged(self, pillow, doc):
        self._configDoc = doc

        if doc != None and (doc.get("rooms." + self._name) == None or not isinstance(doc.get("rooms." + self._name), dict)):
            doc.set("rooms." + self._name, {"Peers": [{"name":"Debugger", "config": {}}]})
            from wallaby.pf.peer.editor import Editor
            House.get("__CONFIG__").throw(Editor.Out.FieldChanged, "rooms")

        self._configChanged(None, "rooms." + self._name)

    def configUpdated(self, pillow, force=False):
        if self._unclean or force:
            self.initialize()
            self.initializePeers()
            self._unclean = False
            self.throwSuggests()

    def throwSuggests(self, pillow=None, feathers=None):
        from wallaby.common.document import Document
        data = {
            "peers": self.suggests(),
            "inBound": sorted(list(self._inBoundPillows)),
            "outBound": sorted(list(self._outBoundPillows)),
            "all": sorted(list(self._inBoundPillows | self._outBoundPillows))
        }

        data['all'].insert(0, "")

        self.throw(Room.Out.Suggests, Document(data=data))

    def _configChanged(self, pillow, path):
        from wallaby.pf.peer.viewer import Viewer
        if not self._configDoc or not Viewer.matchPath("rooms." + self._name, path): return
        config = self._configDoc.get("rooms." + self._name)

        import json
        if config != None and (self._lastConfig == None or json.dumps(config) != json.dumps(self._lastConfig)):
            self._lastConfig = copy.deepcopy(config)
            if 'Peers' in config:
                del self._peers[:] # clear all pre-configured peers
            else: 
                self._peers = []
                config['Peers'] = []

            for mod in config['Peers']:
                if isinstance(mod, dict):
                    self._peers.append(mod)

            self._unclean = True
            self.createPeers()

    def _ignore(self, pillow, feathers):
        pass

    def customPeers(self):
        pass

    def rooms(self, *args):
        rooms = OrderedSet()
        rooms.add(self._name)
        return rooms

    def initializePeers(self):
        done = set()
        for p in self._dynamicPeers:
            p.initialize()
            done.add(p)

        for p in self._allPeers:
            if p not in done:
                p.initialize()

    def initialize(self):
        self.customPeers()
        self.initialized = True

    def createPeers(self):
        for p in self._dynamicPeers: p.destroy(remove=True)

        self._dynamicPeers = []

        observer = House.observer()

        for dct in self._peers:
            name = dct['name']
            if name is None or name == "None": continue

            args = {}
            if 'config' in dct: args = dct['config']

            if not args:
                try:
                    obj = observer.peer(name, self._name)
                except Exception as e:
                    FX.crit("Exception while creating instance of", name)
                    import traceback 
                    traceback.print_exc(file=sys.stdout)

            elif isinstance(args, dict):
                lst = ()

                obj = observer.peer(name, self._name, *lst, **args)

            if not obj:
                FX.crit("Error creating peer", name, "for room", self._name, "with options", args)
            else:
                FX.info("Successfully created peer", name, "for room", self._name)
                self._dynamicPeers.append(obj)
                
    def uncatch(self, pillow, catcher, passThrower=False):
        if ":" in str(pillow):
            parts = pillow.split(':')
            return House.uncatch(parts[0], parts[1], catcher, passThrower)

        if pillow in self._pillows:
            if (catcher, passThrower) in self._pillows[pillow]:
                if Room.registerType == OrderedDict:
                    del self._pillows[pillow][(catcher, passThrower)]
                elif Room.registerType == list:
                    self._pillows[pillow].remove((catcher, passThrower))
                else:
                    self._pillows[pillow].remove((catcher, passThrower))

        if pillow in self._statefullQueue:
            if (catcher, passThrower) in self._statefullQueue[pillow]:
                self._statefullQueue[pillow].remove( (catcher, passThrower) )

    def catchNow(self, pillow):
        d = defer.Deferred()
        self.catch(pillow, partial(self._catchNow, d), single=True)
        return d

    def _catchNow(self, d, pillow, feathers):
        d.callback(feathers)

    def catch(self, pillow, catcher, passThrower=False, single=False):
        if ":" in str(pillow):
            parts = pillow.split(':')
            if parts[0] != self._name:
                return House.catch(parts[0], parts[1], catcher, passThrower, single)

            pillow = unicode(parts[1])

        pillow = unicode(pillow)

        if not single:
            register = self._pillows
        else:
            register = self._singlePillows 

        if not pillow in register:
            register[pillow] = Room.registerType()

        if Room.registerType == OrderedDict:
            register[pillow][(catcher, passThrower)] = True
        elif Room.registerType == list:
            register[pillow].append((catcher, passThrower))
        else:
            register[pillow].add((catcher, passThrower))

        if pillow in self._statefullPillows:
            self.__enqueueStatefull(catcher, pillow, passThrower)
            self.__dequeueRequest()

    def __dequeueRequest(self):
        if self._dequeueCall == None:
            from twisted.internet import reactor
            self._dequeueCall = reactor.callLater(0, self.__dequeue)

    def __enqueueStatefull(self, catcher, pillow, passThrower):
        if pillow not in self._statefullQueue: self._statefullQueue[pillow] = []
        self._statefullQueue[pillow].append( (catcher, passThrower) )

    def __enqueue(self, catcher, pillow, feathers, passThrower=False, **ka):
        if pillow.endswith("!"):
            self.__enqueueStatefull(catcher, pillow, passThrower)
        else:
            self._normalQueue.append( (catcher, (pillow, feathers), ka) ) 

    def __dequeue(self):
        calls = []
        statefullCnt = 0

        for pillow in self._statefullQueue:
            if pillow not in self._statefullPillows: continue

            for ident, (args, ka) in self._statefullPillows[pillow].items():
                statefullCnt += len(self._statefullQueue[pillow])

                for catcher, passThrower in self._statefullQueue[pillow]:
                    if passThrower and 'me' in ka:
                        call = (catcher, args, {"thrower":ka['me']})
                    else:
                        call = (catcher, args, {})

                    calls.append(call)

        self._statefullQueue = {}

        for call in self._normalQueue:
            calls.append(call)

        # print "[", self._name, "] Queue sizes:", statefullCnt, len(self._normalQueue) 

        self._normalQueue = []

        for catcher, args, ka in calls: 
            try:
                catcher(*args, **ka)
            except Exception as e:
                FX.crit("Exception while throwing pillow")
                import traceback 
                traceback.print_exc(file=sys.stdout)

        self._dequeueCall = None
        if len(self._statefullQueue) + len(self._normalQueue) > 0: self.__dequeueRequest()
        
    def enter(self, peer):
        if not peer:
            return

        self._outBoundPillows = self._outBoundPillows | set(peer.dynamicPillows())

        self._allPeers.add(peer)

        peer = peer.__module__ + '.' + peer.__class__.__name__

        if peer in self._loadedPeers:
            return

        self._loadedPeers.append(peer)
        observer = House.observer()

        self._inBoundPillows = self._inBoundPillows | set(observer.inBoundPillows(peer))
        self._outBoundPillows = self._outBoundPillows | set(observer.outBoundPillows(peer))

    def leave(self, peer):
        if not peer:
            return

        if peer in self._allPeers: self._allPeers.remove(peer)

        peer = peer.__module__ + '.' + peer.__class__.__name__

        if peer not in self._loadedPeers:
            return

        observer = House.observer()

        for pillow in observer.outBoundPillows(peer):
            if pillow in self._outBoundPillows:
                self._outBoundPillows.remove(pillow)

        for pillow in observer.inBoundPillows(peer):
            if pillow in self._inBoundPillows:
                self._inBoundPillows.remove(pillow)

        self._loadedPeers.remove(peer)

    def inBoundPillows(self):
        return self._inBoundPillows

    def outBoundPillows(self):
        return self._outBoundPillows

    def allPeers(self):
        return self._allPeers

    def peers(self):
        return self._loadedPeers

    def destroyPeers(self):
        peers = []
        for p in self._allPeers:
            peers.append(p)
        for p in peers:
            p.destroy(remove=True) 

    def throw(self, pillow, *args, **ka): #TODO: refactor to deferredThrow?
        if ":" in str(pillow):
            House.throw(pillow, *args, **ka)
        else:
            self._throw(pillow, *args, **ka)

    def _throw(self, pillow, feathers="", **ka):
        try:
            wcPeers     = None
            normalPeers = None
            singlePeers = None

            if pillow == None: 
                # print "WARNING - skipping none pillow"
                return

            # decode json feathers
            if feathers != None and isinstance(feathers, (str, unicode)) and feathers.startswith('json:'):
                jsonString = '{"object":' + feathers.replace('json:', '', 1) + '}'
                try: 
                    import json
                    feathers = json.loads(jsonString)
                    feathers = feathers["object"]
                except Exception as e:
                    print "Error while parsing json", jsonString, e
                    feathers = None

            # save statefull pillow for resending
            if pillow.endswith("!"):
                identifier = "__any__"
                try: identifier = feathers.identifier()
                except: pass

                if pillow not in self._statefullPillows: self._statefullPillows[pillow] = {}

                self._statefullPillows[pillow][identifier] = ((pillow, feathers), ka)
                self._statefullQueue[pillow] = []

            # create copies of peers to avoid race conditions
            # FIXME: this can result in calls of unallocated QObject methods!!! 
            if "*" in self._pillows:
                wcPeers = copy.copy(self._pillows["*"])

            if pillow in self._pillows:
                normalPeers = copy.copy(self._pillows[pillow])

            if pillow in self._singlePillows:
                singlePeers = copy.copy(self._singlePillows[pillow])
                del self._singlePillows[pillow]

            for peers in (wcPeers, normalPeers, singlePeers):
                if peers != None and pillow != Room.Out.NoHandlerFound:
                    for (catcher, passThrower) in peers:
                        if passThrower and 'me' in ka:
                            self.__enqueue(catcher, pillow, feathers, passThrower=passThrower, thrower=ka['me'])
                        else:
                            self.__enqueue(catcher, pillow, feathers, passThrower=passThrower)

            # Send a "NoHandlerFound" pillow if not already handled
            if normalPeers == None and singlePeers == None and pillow != Room.Out.NoHandlerFound:
                self._throw(Room.Out.NoHandlerFound, pillow, thrower=self)

        except Exception as e:
            FX.crit("Exception while throwing pillow", pillow)
            import traceback 
            traceback.print_exc(file=sys.stdout)

        self.__dequeueRequest()

    def catchers(self, pillow):
        if ":" in str(pillow):
            room,pillow = pillow.split(':')
                
            return House.get(room).catchers(pillow)
        else:
            return self._catchers(pillow)

    def _catchers(self, pillow):
        catchers = []
        if pillow in self._pillows:
            for (catcher, thrower) in self._pillows[pillow]:
                catchers.append(catcher.im_self)

        return catchers 

    def suggests(self):
        observer = House.observer()
        suggests = {}

        for fqa in self._inBoundPillows:
            if fqa not in self._outBoundPillows:
                peers = observer.outBoundPeers(fqa, True)

                # print "WARNING - pillow not thrown:", fqa

                runtimeObject = False
                
                for h in peers:
                    if observer.objectType(h) == Pillow.Runtime:
                        runtimeObject = True

                if runtimeObject:
                    continue

                try:
                    lst = observer.outBoundPeers(fqa, fq=True)
                    for peer in lst:
                        if peer not in suggests:
                            suggests[peer] = 1
                        else:
                            suggests[peer] += 1
                except: 
                    pass

        for peer in self._allPeers:
            try:
                for s in peer.Suggests:
                    if s not in suggests:
                        suggests[s] = 1
                    else:
                        suggests[s] += 1
            except:
                pass

        for peer in self._allPeers:
            if peer.__class__.__name__ in suggests:
                del suggests[peer.__class__.__name__]

        # for fqa in self._outBoundPillows:
        #     if fqa not in self._inBoundPillows:
        #         peers = observer.inBoundPeers(fqa, True)

        #         runtimeObject = False
        #         
        #         for h in peers:
        #             if observer.objectType(h) == Pillow.Runtime:
        #                 runtimeObject = True

        #         if runtimeObject:
        #             continue

        #         try:
        #             lst = observer.intBoundPeers(fqa, fq=True)
        #             suggests = suggests | set(lst)
        #         except: 
        #             pass

        ret = []
        suggests = sorted(suggests.items(), key=itemgetter(1), reverse=True)

        for peer, value in suggests:
            try:
                lst = peer.split('.')
                peer = lst[-1]
                cls = observer.peerClass(peer)

                description = ""
                config = ""

                try:
                    description = cls.Description
                    if isinstance(description, (list, tuple)):
                        description, config = description
                except: pass

                ret.append({"peer":peer, "description": description, "config": config})
            except:
                pass

        return ret

    def check(self, suggest=False):
        print "Checking room", self._name
        observer = House.observer()

        print self.suggests()

        for fqa in self._inBoundPillows:
            if fqa not in self._outBoundPillows:
                peers = observer.outBoundPeers(fqa, True)

                runtimeObject = False
                
                for h in peers:
                    if observer.objectType(h) == Pillow.Runtime:
                        runtimeObject = True

                if runtimeObject:
                    continue

                print "WARNING - pillow not thrown:", fqa
                if suggest:
                    try:
                        pillow = fqa.split('.')[2]
                        lst = observer.outBoundPeers(fqa, fq=True)
                        if lst: print "\tPossible thowers:", lst
                    except:
                        print "ERROR - split:", pillow

        for fqa in self._outBoundPillows:
            if fqa not in self._inBoundPillows:
                peers = observer.inBoundPeers(fqa, True)

                runtimeObject = False
                
                for h in peers:
                    if observer.objectType(h) == Pillow.Runtime:
                        runtimeObject = True

                if runtimeObject:
                    continue

                print "WARNING - pillow not catched:", fqa
                if suggest:
                    try:
                        pillow = fqa.split('.')[2]
                        lst = observer.inBoundPeers(fqa, fq=True)
                        if lst: print "\tPossible catchers:", lst
                    except:
                        print "ERROR - split:", pillow

        for fqa in self._inBoundPillows:
            if fqa not in self._pillows and fqa + "!" not in self._pillows:
                print "WARNING - pillow not registered:", fqa

    @FX.Property
    def registeredPillows():
        doc = "The registeredPillows property"
        def fget(self):
            return self._pillows




# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from directEditDocument import DirectEditDocument

from editor import Editor

class TemplateDocument(DirectEditDocument):

    Sending = [
        Editor.Out.FieldChanged
    ]

    def __init__(self, room, id=None, path=None, proto=None):
        DirectEditDocument.__init__(self, room, id)
        
        self._editDoc = None
        self._path = path
        self._proto = proto

        from viewer import Viewer
        self._catch(Viewer.In.Document, self._editDocChanged)

    def _editDocChanged(self, pillow, doc):
        self._editDoc = doc
        self._id = None

    def _enable(self, enabled):
        pass

    def _saveAs(self, pillow, newID):
        if self._doc == None:
            newDoc = self._editDoc.clone()
            newDoc.resetDocumentID(newID)
            newDoc.overwrite(None, self._proto)

            from documentCache import DocumentCache
            self._throw(DocumentCache.In.CreateAndSaveDocument, newDoc._data)
            return

        self._doc.set(self._path, self._editDoc.get(self._path))
        DirectEditDocument._saveAs(self, pillow, newID)

    def _updateAll(self):
        if self._editDoc is None or self._doc is None: return

        if self._path is not None:
            self._editDoc.set(self._path, self._doc.get(self._path))

            path = Editor.translatePath(self._editDoc, self._path)
            self._throw(Editor.Out.FieldChanged, path)
        else:
            id = self._editDoc.get('_id')
            rev = self._editDoc.get('_rev')

            self._editDoc._data = copy.deepcopy(self._doc._data)
            self._editDoc.set('_id', id)
            if rev is not None: self._editDoc.set('_rev', rev)

            self._throw(Editor.Out.FieldChanged, "")

    def initialize(self):
        pass



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from documentCache import DocumentCache
from viewer import Viewer

class Pref(Peer):
    Load = Pillow.In
    Add  = Pillow.In

    NoDocument  = Pillow.Out
    SheetNotFound  = Pillow.Out
    Ready = Pillow.OutState

    Receiving = [
        Viewer.In.Document,
        DocumentCache.Out.RequestedDocument,
        DocumentCache.Out.DocumentChanged
    ]

    Sending = [
        DocumentCache.In.RequestDocument,
        Viewer.In.Refresh
    ]

    def isReady(self, ignoreVersion=False):
        if self._configDoc is not None and self._document is not None:
            configRev = self._configDoc.rev()
            docRev = self._document.rev()

            if (configRev == None or configRev != self._configRev) or (docRev == None or docRev != self._docRev) or ignoreVersion:
                self._configRev, self._docRev = configRev, docRev

                from twisted.internet import reactor
                reactor.callLater(0, self._controller.load)

            self._throw(Pref.Out.Ready, True)

        else:
            self._throw(Pref.Out.Ready, False)

    def __init__(self, room, controller, configDocId, path):
        Peer.__init__(self, room)
        self._configDoc = None
        self._path = path

        self._configRev = None
        self._docRev = None

        self._document = None

        self._controller = controller
        self._configDocId = configDocId

        self._catch(Pref.In.Add, self._add)
        self._catch(Viewer.In.Document, self._doc)
        self._catch(DocumentCache.Out.RequestedDocument, self._setConfigDoc)
        self._catch(DocumentCache.Out.DocumentChanged, self._setConfigDoc)

    def updateAll(self):
        self._throw(Viewer.In.Refresh, self._path + ".")

    def _setConfigDoc(self, pillow, doc):
        if not doc or doc.documentID != self._configDocId: return
        self._configDoc = doc
        self.isReady()

    def initialize(self):
        self._throw(DocumentCache.In.RequestDocument, self._configDocId)
        self.isReady()

    def noDocument(self):
        self._throw(Pref.Out.NoDocument, None)

    def notFound(self, name):
        self._throw(Pref.Out.SheetNotFound, name)

    def _doc(self, pillow, doc):
        self._document = doc
        self.isReady()

    def document(self):
        return self._document

    def _add(self, pillow, sheet):
        from twisted.internet import reactor
        reactor.callLater(0, self._controller.createSheet, sheet)

    def configDoc(self):
        return self._configDoc



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from twisted.internet import defer, reactor
from twisted.trial import unittest

from wallaby.pf.room import House
from wallaby.pf.peer.peer import Peer
from wallaby.pf.peer.credentialsParty import CredentialsParty

factory = House.observer()

def sleep(time):
    d = defer.Deferred()
    reactor.callLater(time, d.callback, None)
    return d

class WallabyCouchDBTest(unittest.TestCase):
    def setUp(self):
        self._dbName = "wallaby_test"
        self._docId  = "testdoc"
        self.timeout = 1

        self._designDoc = {
           "_id": "_design/wallaby_test",
           "language": "javascript",
           "views": {
               "text": {
                   "map": "function(doc) { if(doc.text)\n   emit(doc.text, null);\n}"
               }
           },
           "filters": {
               "typeB": "function(doc, req) { if((doc.type && doc.type == \"typeB\") || doc.deleted) { return true; } return false;}"
           }
        }

        import wallaby.backends.couchdb as couch
        self._db = couch.Database.getDatabase(self._dbName)

    def tearDown(self):
        import wallaby.backends.couchdb as couch
        couch.Database.closeDatabase(self._dbName)

    @defer.inlineCallbacks
    def getDoc(self, name):
        doc = yield self._db.get(name)
        self.assertTrue(doc != None)
        self.assertEqual(doc["_id"], name)
        defer.returnValue(doc)

    @defer.inlineCallbacks
    def test_00_initialize(self):
        try:
            info = yield self._db.info(keepOnTrying=False, returnOnError=True)
            if info != None: 
                yield self._db.destroy()
        except:
            pass

        ret = yield self._db.create()
        self.assertTrue(ret["ok"])
        yield self._db.save(self._designDoc)
        yield self._db.save({"_id": self._docId, "text":"Hello World!"})
        yield self._db.save({"_id": "credentials"})

    @defer.inlineCallbacks
    def test_01_database(self):
        global factory
        peers = ["CouchDB"]

        for p in peers:
            factory.peer(p, "room01")

        House.initializeAll()
        Peer.initializeAll()
        yield sleep(0.1)

        d = defer.Deferred()

        room = House.get("room01")

        from wallaby.pf.peer.database import Database
        room.catch(Database.Out.RequestedDocument, lambda *args: d.callback(args))
        room.throw(Database.In.RequestDocument, self._docId)

        pillow, doc = yield d
        self.assertEqual(doc.documentID, self._docId)

        House.destroyRoom("room01")

    @defer.inlineCallbacks
    def test_02_view(self):
        global factory
        peers = ["ViewDocument", "DocumentCache", "DocumentDatabase", "CouchDB"]

        for p in peers:
            factory.peer(p, "room01")

        CredentialsParty()

        d = defer.Deferred()
        def callback(value):
            if value:
                d.callback(value)

        from wallaby.pf.peer.viewer import Viewer
        Viewer("room01", callback, "text")

        House.initializeAll()
        Peer.initializeAll()
        yield sleep(0.1)

        from wallaby.pf.peer.editDocument import EditDocument
        room = House.get("room01")
        room.throw(EditDocument.In.Load, self._docId)

        value = yield d
        self.assertEqual(value, "Hello World!")

        House.destroyRoom("room01")
        House.destroyRoom("__CREDENTIALS__")

    @defer.inlineCallbacks
    def test_03_edit(self):
        global factory
        peers = ["EditDocument", "DocumentCache", "DocumentDatabase", "CouchDB"]

        for p in peers:
            factory.peer(p, "room01")

        CredentialsParty()

        valueDeferred = defer.Deferred()
        def valueCallback(value):
            if value == "Hello Jan": 
                try:
                    valueDeferred.callback(value)
                except: pass

        setRODeferred = defer.Deferred()
        def roCallback(readOnly):
            if not readOnly: 
                try:
                    setRODeferred.callback(readOnly)
                except: pass

        from wallaby.pf.peer.viewer import Viewer
        from wallaby.pf.peer.editor import Editor
        Viewer("room01", valueCallback, "text")
        editor = Editor("room01", path="text", setReadOnlyCallback=roCallback)

        House.initializeAll()
        Peer.initializeAll()
        yield sleep(0.1)

        from wallaby.pf.peer.editDocument import EditDocument
        room = House.get("room01")
        room.throw(EditDocument.In.LoadAndEdit, self._docId)
        
        yield setRODeferred
        editor.changeValue("Hello Jan")

        value = yield valueDeferred
        self.assertEqual(value, "Hello Jan")

        room.throw(EditDocument.In.Save, self._docId)

        from wallaby.pf.peer.database import Database
        docId = yield room.catchNow(Database.Out.DocumentSaved)
        self.assertEqual(docId, self._docId)

        doc = yield self._db.get(self._docId)
        self.assertEqual(doc["text"], "Hello Jan")

        House.destroyRoom("room01")
        House.destroyRoom("__CREDENTIALS__")

    @defer.inlineCallbacks
    def test_99_destroy(self):
        res = yield self._db.destroy()
        self.assertTrue(res["ok"])



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

import twisted.application.service
from twisted.python import logfile
import wallaby.FX as FX
import twisted.python.log as log
import os

class FXLogger(twisted.application.service.Service):
    def __init__(self, logName=None, logDir="./log/"):
        self.logName = logName
        self.logDir = logDir
        self.maxLogSize = 1000000
        self.pidfile = None

    def setPIDFile(self, pidfile):
        self.pidfile = pidfile
       
    def startService(self):
        if self.logName and os.path.exists(self.logDir):
            self.logFile = logfile.LogFile(self.logName, self.logDir, rotateLength=self.maxLogSize)
            self.errlog = log.FileLogObserver(self.logFile)
            self.errlog.start(  )

        if self.pidfile:
	    # self.pidfile = 'c:/test2.pid'
            print "Writing pid", os.getpid(), "to pidfile:", (">" + self.pidfile + "<")
            f = open(self.pidfile, "w+")
            f.write(str(os.getpid()))
            f.close()
        
    def stopService(self):
        if self.logName:
            self.errlog.stop(  )
            self.logFile.close(  )

	if self.pidfile:
            if os.path.isfile(self.pidfile):
                os.remove(self.pidfile)






# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from multiViewer import MultiViewer
from viewer import Viewer
from editor import Editor

class MultiViewEditor(Peer):
    Description = ("Routing between editor elements and the query document of a multi-view element (like a table)",
                    '{ "dstRoom": null, "pillow": "MultiViewer.In.Refresh" }')

    Receiving = [
        MultiViewer.Out.QueryDocument,
        Editor.Out.FieldChanged
    ] 

    Sending = [
        Editor.In.Enable
    ]

    Routings = [
        (MultiViewer.Out.QueryDocument, Viewer.In.Document),
        (Editor.Out.FieldChanged, Viewer.In.Refresh)
    ]

    def __init__(self, room, dstRoom=None, pillow="MultiViewer.In.Refresh"):
        Peer.__init__(self, room)
       
        self._dstRoom = dstRoom
        self._pillow = pillow

        self._catch(MultiViewer.Out.QueryDocument, self._enable)

        if dstRoom is not None:
            self._catch(Editor.Out.FieldChanged, self._fieldChanged)

    def _fieldChanged(self, pillow, feathers):
        if self._dstRoom:
            self._throw(self._dstRoom + ":" + self._pillow, None)

    def _enable(self, pillow, feathers):
        self._throw(Editor.In.Enable, True)

    def initialize(self):
        pass 






# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class EditDocumentDialog(Peer):
    def __init__(self, room, database, id, dialog, dbClass=None, *args):
        from directEditDocument import DirectEditDocument
        from documentDatabase import DocumentDatabase
        from debugger import Debugger
        from dialog import Dialog

        Peer.__init__(self, room)
       
        self.add(DirectEditDocument(room, id))
        self.add(DocumentDatabase(room))

        self.add(Debugger(room, "*"))
        self.add(DocumentCache(room))
        self.add(Dialog(room, dialog))

        if dbClass is None:
            from couchDB import CouchDB
            self.add(CouchDB(room, queryPeer=False, databaseName=database))
        else:
            self.add(dbClass(room, queryPeer=False, databaseName=database))



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from wallaby.common.document import Document

class RoomDebugger(Peer):

    RequestRooms  = Pillow.In
    RequestDebugInfo = Pillow.In

    RequestedRooms  = Pillow.Out
    RequestedDebugInfo = Pillow.Out

    def __init__(self, room):
        Peer.__init__(self, room)
        self._roomName = room
        self._catch(ContextDebugger.In.RequestRooms, self._rooms)
        self._catch(ContextDebugger.In.RequestDebugInfo, self._info)

    def _rooms(self, pillow, feathers):
        dict = {'_id':'rooms'}
        rooms = House.rooms.keys()
        rooms.sort()

        dict['titles'] = rooms
        document = Document(data=dict)
        self._throw(ContextDebugger.Out.RequestedRooms, document)

    def _info(self, pillow, feathers):
        # Get registered pillows and feathers
        pillows = None
        dict = {'_id':'pillows','room':feathers}

        if feathers in House.rooms:
            pillows = House.rooms[feathers].registeredPillows
            inBoundPillows = House.rooms[feathers].inBoundPillows()
            outBoundPillows = House.rooms[feathers].outBoundPillows()
            peers = House.rooms[feathers].peers()
        else:
            pillows = House.rooms[self._roomName].registeredPillows
            inBoundPillows = House.rooms[self._roomName].inBoundPillows()
            outBoundPillows = House.rooms[self._roomName].outBoundPillows()
            peers = House.rooms[self._roomName].peers()
            dict['room'] = self._roomName

        pillowKeys = pillows.keys()
        pillowKeys.sort()

        pillowDicts = []

        for pillow in pillowKeys:
            callbackDict = {}
            callbacks = pillows[pillow]
            for callback,_ in callbacks:
                className = callback.im_self.__class__.__name__
                functionName = callback.im_func.__name__
                if className in callbackDict and callbackDict[className]['function'] == functionName:
                    callbackDict[className]['count'] += 1
                else:
                    callbackDict[className] = {'class':callback.im_self.__class__.__name__,'function':callback.im_func.__name__,'count':1}

            keys = callbackDict.keys()
            keys.sort()
            callbackDicts = []
            for key in keys:
                callbackDicts.append(callbackDict[key])

            pillowDicts.append({'name':pillow,'callbacks':callbackDicts})

        dict['pillows'] = pillowDicts

        # Get unhandled and uncalled pillows
        observer = House.observer()

        # Uncalled pillows
        uncalled = []
        for fqa in inBoundPillows:
            if fqa not in outBoundPillows:
                peers = observer.outBoundPeers(fqa, True)
                
                runtimeObject = False

                for p in peers:
                    if observer.objectType(p) == Pillow.Runtime:
                        runtimeObject = True

                if runtimeObject:
                    continue

                uc = {'pillow': fqa, 'suggests': []}

                try:
                    plw = fqa.split('.')[2]
                    lst = observer.outBoundPeers(plw)
                    if lst: uc['suggests'] =  lst
                except:
                    print "ERROR - split:", plw

                uncalled.append(uc)

        dict['uncalled'] = uncalled

        # Unhandled pillows
        unhandled = []
        for fqa in outBoundPillows:
            if fqa not in inBoundPillows:
                peers = observer.inBoundPeers(fqa, True)
                
                runtimeObject = False

                for p in peers:
                    if magic.objectType(p) == Pillow.Runtime:
                        runtimeObject = True

                if runtimeObject:
                    continue

                uh = {'pillow': fqa, 'suggests': []}

                try:
                    plw = fqa.split('.')[2]
                    lst = observer.inBoundPeers(act)
                    if lst: uh['suggests'] = lst
                except:
                    print "ERROR - split:", act

                uncalled.append(uc)

        dict['unhandled'] = unhandled

        unregistered = []
        for fqa in inBoundPillows:
            if fqa not in peers:
                unregistered.append({'pillow': fqa})    

        dict['unregistered'] = unregistered

        document = Document(data=dict)
        self._throw(ContextDebugger.Out.RequestedDebugInfo, document)
